# Copyright 2024 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Types and constructs for PyTree metadata and serialization."""

from __future__ import annotations

import abc
import dataclasses
from typing import Any, Callable, Optional, Protocol, Sequence, Tuple

from absl import logging
from etils import epath
import jax
import jax.numpy as jnp
import numpy as np
from orbax.checkpoint import future
from orbax.checkpoint._src.metadata import empty_values
from orbax.checkpoint._src.metadata import value as value_metadata
from orbax.checkpoint._src.serialization import serialization
import tensorstore as ts


def is_supported_type(value: Any) -> bool:
  """Determines if the `value` is supported without custom TypeHandler."""
  return isinstance(
      value,
      (str, int, float, np.number, np.ndarray, bytes, jax.Array),
  ) or empty_values.is_supported_empty_value(value)


def get_param_typestr(value: Any, registry: TypeHandlerRegistry) -> str:
  """Retrieves the typestr for a given value."""
  if empty_values.is_supported_empty_value(value):
    typestr = empty_values.get_empty_value_typestr(value)
  else:
    try:
      handler = registry.get(type(value))
      typestr = handler.typestr()
    except ValueError:
      # Not an error because users' training states often have a bunch of
      # random unserializable objects in them (empty states, optimizer
      # objects, etc.). An error occurring due to a missing TypeHandler
      # will be surfaced elsewhere.
      typestr = empty_values.RESTORE_TYPE_NONE
  return typestr


@dataclasses.dataclass
class ParamInfo:
  """Information describing a parameter in a PyTree.

  Note that ParamInfo is distinct from SaveArgs and RestoreArgs in that in
  represents information not provided by a user, and should be computed
  internally.

  name:
    Name of the parameter.
  path:
    A path providing a location where file(s) should be saved. The path is
    assumed to be a directory.
  parent_dir:
    A path providing location where all files under the same checkpoint should
    be saved under. All `ParamInfo` provided to a given TypeHandler should have
    the same `parent_dir`. The parent_dir is assumed to be a directory.
  skip_deserialize:
    If specified, skips deserialization of the given parameter using the
    TypeHandler. This may be for multiple different reasons, including that the
    parameter may have been aggregated, or it will be unneeded after
    transformations. Note: this parameter is handled by PyTreeCheckpointHandler,
    so it is unnecessary for TypeHandler implementations to deal with it.
  byte_limiter:
    Object to limit the number of bytes that can be read in
    parallel.
  is_ocdbt_checkpoint:
    Indicates whether the checkpoint path uses OCDBT format
    or not. Only used for restoration.
  use_zarr3:
    If True, use Zarr ver3 otherwise ver2.
  ocdbt_target_data_file_size:
    Specifies the target size (in bytes) of each OCDBT data file. If set to 0,
    data file size is not limited. If omitted (None), the TensorStore default
    is used.
  ts_context:
    Tensorstore context to use for reading/writing.
  value_typestr: stores the original value's typestr (from TypeHandler).
    Only required when saving.
  enable_pinned_host_transfer:
    True by default. If False, disables transfer to pinned host when copying
    from device to host, regardless of the presence of pinned host memory.
  raise_array_data_missing_error:
    Only used for restoring. See documentation in `tensorstore_utils.py`. Comes
    from tree metadata and should be the same across all parameters.
  """

  name: Optional[str] = None
  path: Optional[epath.Path] = None
  parent_dir: Optional[epath.Path] = None
  skip_deserialize: Optional[bool] = None
  byte_limiter: Optional[serialization.LimitInFlightBytes] = None
  is_ocdbt_checkpoint: Optional[bool] = None
  use_zarr3: Optional[bool] = False
  ocdbt_target_data_file_size: Optional[int] = None
  ts_context: Optional[ts.Context] = None
  value_typestr: Optional[str] = None
  enable_pinned_host_transfer: bool = True
  raise_array_data_missing_error: bool = True


@dataclasses.dataclass
class SaveArgs:
  """Extra arguments that can be provided for saving.

  aggregate:
    Deprecated, please use custom TypeHandler
    (https://orbax.readthedocs.io/en/latest/custom_handlers.html#typehandler) or
    contact Orbax team to migrate before August 1st, 2024. If true, saves the
    given
    parameter in an aggregated tree format rather than individually. See
    AggregateHandler.
  dtype:
    If provided, casts the parameter to the given dtype before saving.
    Note that the parameter must be compatible with the given type (e.g.
    jnp.bfloat16 is not compatible with np.ndarray).
  chunk_byte_size:
    This is an experimental feature that automatically chooses the largest chunk
    shape possible, while keeping the chunk byte size less than or equal to the
    specified chunk_byte_size. Both the write_chunk_shape and read_chunk_shape
    are automatically set to the chosen shape. This uses a greedy algorithm that
    prioritizes splitting the largest dimensions first.
  """

  aggregate: bool = False
  dtype: Optional[jnp.dtype] = None
  chunk_byte_size: Optional[int] = None

  def __post_init__(self):
    if self.aggregate:
      jax.monitoring.record_event('/jax/orbax/deprecation/aggregate')
      logging.log_every_n_seconds(
          logging.WARNING,
          'The `aggregate` option is deprecated and will be ignored.',
          n_seconds=12 * 60 * 60,  # once every 12 hours
      )


@dataclasses.dataclass
class RestoreArgs:
  """Extra arguments that can be provided for restoration.

  restore_type:
    Specifies the object type of the restored parameter. The type
    must have a corresponding TypeHandler for restoration. Ignored if the
    parameter is restored from an aggregated checkpoint file.
  dtype:
    If provided, casts the parameter to the given dtype after restoring.
    Note that the parameter must be compatible with the given type (e.g.
    jnp.bfloat16 is not compatible with np.ndarray).
  """

  restore_type: Optional[Any] = None
  dtype: Optional[jnp.dtype] = None


class TypeHandler(abc.ABC):
  """Interface for reading and writing a PyTree leaf."""

  @abc.abstractmethod
  def typestr(self) -> str:
    """A string representation of the type.

    Cannot conflict with other types.

    Returns:
      The type as a string.
    """
    pass

  @abc.abstractmethod
  async def metadata(
      self, infos: Sequence[ParamInfo]
  ) -> Sequence[value_metadata.Metadata]:
    """Constructs object metadata from a stored parameter location.

    Args:
      infos: sequence of ParamInfo

    Returns:
      Sequence of Metadata for each provided ParamInfo.
    """
    pass

  @abc.abstractmethod
  async def serialize(
      self,
      values: Sequence[Any],
      infos: Sequence[ParamInfo],
      args: Optional[Sequence[SaveArgs]] = None,
  ) -> Sequence[future.Future]:
    """Writes the parameter to a storage location.

    This method is responsible for copying the parameter from a remote device in
    a synchronous fashion (if applicable). It should then return a list of
    futures which can be later awaited to complete the final commit operation
    to a storage location.

    The function can be used in a multihost setting, but should not implement
    extra logic to ensure atomicity.

    Args:
      values: a sequence of parameters to save.
      infos: a sequence of ParamInfo containing relevant information for
        serialization of each value.
      args: a sequence of additional arguments for serialization, provided by
        the user.

    Returns:
      Sequence of commit futures which can be awaited to complete the save
      operation.
    """
    pass

  @abc.abstractmethod
  async def deserialize(
      self,
      infos: Sequence[ParamInfo],
      args: Optional[Sequence[RestoreArgs]] = None,
  ) -> Sequence[Any]:
    """Reads the parameter from a storage location.

    Args:
      infos: Sequence of ParamInfo for deserialization.
      args: Sequence of user-provided restoration information.

    Returns:
      The deserialized parameters.
    """
    pass

  def finalize(self, directory: epath.Path):
    """Performs any logic to finalize parameter files written by this class.

    By default, does nothing.

    Args:
      directory: A path to the location of the checkpoint. This corresponds to
        `param_info.parent_dir`.
    """
    pass

  def memory_size(self, values: Sequence[Any]) -> Sequence[Tuple[int, int]]:
    """For a batch of values, returns the size of each value in bytes.

    Note that the default implementation uses `sys.getsizeof`, which is not
    likely to be accurate for many types.

    The value returned is intended to be per-host.

    Args:
      values: A batch of values.

    Returns:
      A sequence of elements corresponding to `values`. Each element is a tuple
      of [write_size, read_size]. In many cases these values may be the same.

    Raises:
      NotImplementedError: Raises error by default since we will rely on a
        backup implementation.
    """
    raise NotImplementedError()


class TypeHandlerRegistry(Protocol):
  """A registry for TypeHandlers.

  This internal base class is used for the global registry which serves as a
  default for any type not found in a local registry. It is also accessed
  through the module function get/set/has_type_handler.
  """

  def add(
      self,
      ty: Any,
      handler: TypeHandler,
      func: Optional[Callable[[Any], bool]] = None,
      override: bool = False,
  ):
    """Registers a type for serialization/deserialization with a given handler.

    Note that it is possible for a type to match multiple different entries in
    the registry, each with a different handler. In this case, only the first
    match is used.

    Args:
      ty: A type to register.
      handler: a TypeHandler capable of reading and writing parameters of type
        `ty`.
      func: A function that accepts a type and returns True if the type should
        be handled by the provided TypeHandler. If this parameter is not
        specified, defaults to `lambda t: issubclass(t, ty)`.
      override: if True, will override an existing mapping of type to handler.

    Raises:
      ValueError if a type is already registered and override is False.
    """
    ...

  def get(self, ty: Any) -> TypeHandler:
    """Returns the handler registered for a given type, if available.

    Args:
      ty: an object type (or string representation of the type.)

    Returns:
      The TypeHandler that is registered for the given type.

    Raises:
      ValueError if the given type has no registered handler.
    """
    ...

  def has(self, ty: Any) -> bool:
    """Checks if a type is registered.

    Args:
      ty: an object type (or string representation of the type.)

    Returns:
      A boolean indicating if ty is registered.
    """
    ...



# Copyright 2024 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for RandomKeyCheckpointHandler."""

from absl.testing import absltest
from etils import epath
import jax
import numpy as np
from orbax.checkpoint import args as args_lib
from orbax.checkpoint._src.handlers import composite_checkpoint_handler
from orbax.checkpoint._src.handlers import random_key_checkpoint_handler

# Parse absl flags test_srcdir and test_tmpdir.
jax.config.parse_flags_with_absl()

JaxRandomKeyCheckpointHandler = (
    random_key_checkpoint_handler.JaxRandomKeyCheckpointHandler
)

NumpyRandomKeyCheckpointHandler = (
    random_key_checkpoint_handler.NumpyRandomKeyCheckpointHandler
)

CompositeCheckpointHandler = (
    composite_checkpoint_handler.CompositeCheckpointHandler
)


class RandomKeyCheckpointHandlerTest(absltest.TestCase):

  def setUp(self):
    super().setUp()
    self.directory = epath.Path(
        self.create_tempdir(name='checkpointing_test').full_path
    )

  def assert_dict_equal(self, left, right):
    """Asserts that two dicts are equal and allow np.ndarray as elements."""
    assert isinstance(left, type(right))
    for k, v in left.items():
      if k not in right:
        raise ValueError(f'Missing key:{k} in `right`')

      if isinstance(v, dict):
        self.assert_dict_equal(v, right[k])
      elif isinstance(v, np.ndarray):
        np.testing.assert_array_equal(v, right[k])
      else:
        self.assertEqual(v, right[k])

  def assert_tuple_equal(self, left, right):
    """Asserts that two tuples are equal and allow np.ndarray as elements."""
    self.assertIsInstance(left, type(right))
    self.assertLen(left, len(right))

    for i, v in enumerate(left):
      if isinstance(v, np.ndarray):
        np.testing.assert_array_equal(v, right[i])
      else:
        self.assertEqual(v, right[i])

  def test_save_and_restore_jax_random_key_typed(self):
    typed_key = jax.random.key(0)
    handler = JaxRandomKeyCheckpointHandler('typed_key')
    handler.save(self.directory, args_lib.JaxRandomKeySave(item=typed_key))
    handler.finalize(self.directory)

    restore_handler = JaxRandomKeyCheckpointHandler('typed_key')
    restore_typed_keys = restore_handler.restore(
        directory=self.directory, args=args_lib.JaxRandomKeyRestore()
    )
    self.assertTrue(jax.numpy.array_equal(typed_key, restore_typed_keys))

  def test_save_and_restore_jax_random_key_untyped(self):
    typed_key = jax.random.key(0)
    untyped_key = jax.random.PRNGKey(0)
    handler = JaxRandomKeyCheckpointHandler('untyped_key')
    handler.save(
        self.directory,
        args_lib.JaxRandomKeySave(item=untyped_key),
    )
    handler.finalize(self.directory)

    restore_handler = JaxRandomKeyCheckpointHandler('untyped_key')
    restore_untyped_keys = restore_handler.restore(
        directory=self.directory, args=args_lib.JaxRandomKeyRestore()
    )
    self.assertTrue(jax.numpy.array_equal(untyped_key, restore_untyped_keys))

    self.assertFalse(jax.numpy.array_equal(typed_key, untyped_key))

  def test_save_and_restore_numpy_random_key_legacy(self):
    # np random state
    random_state = np.random.get_state()
    handler = NumpyRandomKeyCheckpointHandler('legacy')
    handler.save(self.directory, args_lib.NumpyRandomKeySave(item=random_state))
    handler.finalize(self.directory)

    restore_handler = NumpyRandomKeyCheckpointHandler('legacy')
    restored_random_state = restore_handler.restore(
        directory=self.directory, args=args_lib.NumpyRandomKeyRestore()
    )

    self.assert_tuple_equal(random_state, restored_random_state)

  def test_save_and_restore_numpy_random_key_nonlegacy(self):
    # np random state
    random_state = np.random.get_state(legacy=False)
    handler = NumpyRandomKeyCheckpointHandler('nonlegacy')
    handler.save(
        self.directory, args=args_lib.NumpyRandomKeySave(item=random_state)
    )
    handler.finalize(self.directory)

    restore_handler = NumpyRandomKeyCheckpointHandler('nonlegacy')
    restored_random_state = restore_handler.restore(
        directory=self.directory, args=args_lib.NumpyRandomKeyRestore()
    )

    # dictionary
    self.assert_dict_equal(random_state, restored_random_state)

  def test_save_and_restore_random_keys_with_new_api(self):
    handler = CompositeCheckpointHandler(
        'some_pytree',
        'jax_typed_key',
        'jax_untyped_key',
        'numpy_legacy_key',
        'numpy_nonlegacy_key',
    )

    jax_typed_key = jax.random.key(123)
    jax_untyped_key = jax.random.PRNGKey(256)
    pytree = {'train_state': jax.random.uniform(jax_typed_key, shape=(1, 2, 3))}

    np.random.seed(123)
    np_legacy_key = np.random.get_state(legacy=True)
    np_nonlegacy_key = np.random.get_state(legacy=False)

    handler.save(
        self.directory,
        args=args_lib.Composite(
            some_pytree=args_lib.PyTreeSave(pytree),
            jax_typed_key=args_lib.JaxRandomKeySave(jax_typed_key),
            jax_untyped_key=args_lib.JaxRandomKeySave(jax_untyped_key),
            numpy_legacy_key=args_lib.NumpyRandomKeySave(np_legacy_key),
            numpy_nonlegacy_key=args_lib.NumpyRandomKeySave(np_nonlegacy_key),
        ),
    )

    handler.finalize(self.directory)

    restore_handler = CompositeCheckpointHandler(
        'some_pytree',
        'jax_typed_key',
        'jax_untyped_key',
        'numpy_legacy_key',
        'numpy_nonlegacy_key',
    )

    restored = restore_handler.restore(
        directory=self.directory,
        args=args_lib.Composite(
            some_pytree=args_lib.PyTreeRestore(),
            jax_typed_key=args_lib.JaxRandomKeyRestore(),
            jax_untyped_key=args_lib.JaxRandomKeyRestore(),
            numpy_legacy_key=args_lib.NumpyRandomKeyRestore(),
            numpy_nonlegacy_key=args_lib.NumpyRandomKeyRestore(),
        ),
    )

    self.assertTrue(
        jax.numpy.array_equal(
            pytree['train_state'], restored['some_pytree']['train_state']
        )
    )
    self.assertTrue(
        jax.numpy.array_equal(jax_typed_key, restored['jax_typed_key'])
    )
    self.assertTrue(
        jax.numpy.array_equal(jax_untyped_key, restored['jax_untyped_key'])
    )
    self.assert_tuple_equal(np_legacy_key, restored['numpy_legacy_key'])
    self.assert_dict_equal(np_nonlegacy_key, restored['numpy_nonlegacy_key'])


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test for standard_checkpoint_handler.py."""

import functools
from typing import Any

from absl.testing import parameterized
from etils import epath
import flax
import flax.training.train_state
import jax
from jax import numpy as jnp
from jax.experimental import layout
import numpy as np
import optax
from orbax.checkpoint import test_utils
from orbax.checkpoint import utils
from orbax.checkpoint._src.handlers import standard_checkpoint_handler
from orbax.checkpoint._src.multihost import multihost
from orbax.checkpoint._src.serialization import type_handlers

DLL = layout.DeviceLocalLayout
Layout = layout.Layout
PyTree = Any
SaveArgs = type_handlers.SaveArgs
StandardRestoreArgs = standard_checkpoint_handler.StandardRestoreArgs
StandardSaveArgs = standard_checkpoint_handler.StandardSaveArgs


class StandardCheckpointHandler(
    standard_checkpoint_handler.StandardCheckpointHandler
):

  def save(self, directory, *args, **kwargs):
    super().save(directory, *args, **kwargs)
    test_utils.sync_global_processes('StandardCheckpointHandler:save')
    if multihost.process_index() == 0:
      self.finalize(directory)
    test_utils.sync_global_processes('StandardCheckpointHandler:finalize')


# Not in common util because we need to eliminate OSS dependency on flax.
def init_flax_model(model):
  params = model.init(jax.random.PRNGKey(0), jnp.ones([8, 8]))
  tx = optax.adamw(learning_rate=0.001)
  state = flax.training.train_state.TrainState.create(
      apply_fn=model.apply, params=params, tx=tx
  )
  return jax.tree.map(np.asarray, state)


class StandardCheckpointHandlerTestBase:
  """Base test cases for StandardCheckpointHandler."""

  class Test(parameterized.TestCase):
    """Test class."""

    save_args_cls = StandardSaveArgs
    restore_args_cls = StandardRestoreArgs

    def setUp(self):
      super().setUp()

      self.numpy_pytree = test_utils.setup_pytree()
      pytree, _, _ = test_utils.setup_sharded_pytree(self.numpy_pytree)
      zeros_pytree = jax.tree.map(
          np.zeros_like,
          self.numpy_pytree,
          is_leaf=test_utils.is_leaf,
      )
      zeros_pytree, _, _ = test_utils.setup_sharded_pytree(zeros_pytree)
      self.zeros_pytree = zeros_pytree

      self.numpy_pytree.update({'x': 4.5, 'y': 3})
      self.pytree = pytree
      self.mixed_pytree = {
          'sharded': self.pytree,
          'numpy': self.numpy_pytree,
      }

      self.directory = epath.Path(
          self.create_tempdir(name='checkpointing_test').full_path
      )
      test_utils.set_tensorstore_driver_for_test()

      test_utils.sync_global_processes(
          'StandardCheckpointHandler:setup_complete'
      )

    def tearDown(self):
      test_utils.sync_global_processes(
          'StandardCheckpointHandler:tests_complete'
      )
      self.handler.close()
      super().tearDown()

    @property
    def handler(self) -> StandardCheckpointHandler:
      return StandardCheckpointHandler()

    def test_basic(self):
      self.handler.save(self.directory, args=self.save_args_cls(self.pytree))
      self.assertTrue(type_handlers.is_ocdbt_checkpoint(self.directory))
      self.assertTrue(type_handlers.is_ocdbt_checkpoint(self.directory))
      restored = self.handler.restore(
          self.directory, args=self.restore_args_cls(self.zeros_pytree)
      )
      test_utils.assert_tree_equal(self, self.pytree, restored)

    def test_basic_no_item_arg(self):
      self.handler.save(self.directory, args=self.save_args_cls(self.pytree))
      restored = self.handler.restore(self.directory)
      self.assertTrue(type_handlers.is_ocdbt_checkpoint(self.directory))
      test_utils.assert_tree_equal(self, self.pytree, restored)

    def test_shape_dtype_struct(self):
      """Test case."""
      self.handler.save(
          self.directory, args=self.save_args_cls(self.mixed_pytree)
      )
      restored = self.handler.restore(
          self.directory,
          args=self.restore_args_cls(
              jax.tree.map(utils.to_shape_dtype_struct, self.mixed_pytree)
          ),
      )
      test_utils.assert_tree_equal(self, self.mixed_pytree, restored)

    def test_custom_layout(self):
      """Test custom layout at restoration."""
      inp = np.arange(32).reshape(8, 4)
      mesh = jax.sharding.Mesh(
          np.asarray(jax.devices()).reshape(4, 2), ('x', 'y')
      )
      pspec = jax.sharding.PartitionSpec('x', 'y')
      arr = test_utils.create_sharded_array(inp, mesh, pspec)
      pytree = {'x': arr}

      self.handler.save(self.directory, args=self.save_args_cls(pytree))
      restored_regular = self.handler.restore(
          self.directory, args=self.restore_args_cls(item=pytree)
      )
      test_utils.assert_tree_equal(self, pytree, restored_regular)

      # create a custom layout
      custom_layout = Layout(
          device_local_layout=DLL(
              major_to_minor=arr.layout.device_local_layout.major_to_minor[::-1],  # pytype: disable=attribute-error
              _tiling=arr.layout.device_local_layout._tiling,  # pylint: disable=protected-access  # pytype: disable=attribute-error
          ),
          sharding=arr.sharding,
      )
      arr_new_layout = jax.device_put(arr, custom_layout)
      self.assertNotEqual(arr_new_layout.layout, arr.layout)

      # use a pytree example
      pytree_new_layout = {'x': arr_new_layout}
      restored_new_layout = self.handler.restore(
          self.directory,
          args=self.restore_args_cls(item=pytree_new_layout),
      )
      test_utils.assert_tree_equal(self, pytree_new_layout, restored_new_layout)

      # use shape_dtype_struct with custom layout
      shape_dtypes = jax.tree.map(utils.to_shape_dtype_struct, pytree)
      shape_dtypes = jax.tree.map(
          lambda x: jax.ShapeDtypeStruct(
              x.shape, x.dtype, sharding=custom_layout
          ),
          shape_dtypes,
      )

      restored_shape_dtypes = self.handler.restore(
          self.directory,
          args=self.restore_args_cls(item=shape_dtypes),
      )
      test_utils.assert_tree_equal(
          self, pytree_new_layout, restored_shape_dtypes
      )

    @parameterized.parameters((True,), (False,))
    def test_change_shape(self, strict: bool):
      """Test case."""
      if not hasattr(self.restore_args_cls, 'strict'):
        self.skipTest('strict option not supported for this handler')
      mesh = jax.sharding.Mesh(np.asarray(jax.devices()), ('x',))
      axes = jax.sharding.PartitionSpec(None)
      pytree = {'x': test_utils.create_sharded_array(np.arange(8), mesh, axes)}
      self.handler.save(self.directory, args=self.save_args_cls(pytree))

      mesh = jax.sharding.Mesh(np.asarray(jax.devices()), ('x',))
      axes = jax.sharding.PartitionSpec(None)
      expected_pytree = {
          'x': test_utils.create_sharded_array(np.arange(4), mesh, axes)
      }

      if strict:
        with self.assertRaises(BaseException):
          self.handler.restore(
              self.directory,
              args=self.restore_args_cls(
                  jax.tree.map(utils.to_shape_dtype_struct, expected_pytree),
                  strict=strict,
              ),
          )
      else:
        restored = self.handler.restore(
            self.directory,
            args=self.restore_args_cls(
                jax.tree.map(utils.to_shape_dtype_struct, expected_pytree),
                strict=strict,
            ),
        )
        test_utils.assert_tree_equal(self, expected_pytree, restored)

    def test_save_unsupported_type(self):
      pytree = {'str_key': 'str_value', **self.pytree}
      with self.assertRaisesRegex(ValueError, 'Unsupported type'):
        self.handler.save(self.directory, args=self.save_args_cls(pytree))

    def test_restore_unsupported_type(self):
      pytree = {'str_key': 'str_value', **self.pytree}
      self.handler.save(self.directory, args=self.save_args_cls(self.pytree))
      with self.assertRaisesRegex(ValueError, 'Unsupported type'):
        self.handler.restore(self.directory, args=self.restore_args_cls(pytree))

    def test_cast(self):
      """Test case."""
      # TODO(dicentra): casting from int dtypes currently doesn't work
      # in the model surgery context.
      save_args = jax.tree.map(
          lambda _: SaveArgs(dtype=jnp.dtype(jnp.float32)), self.pytree
      )
      self.handler.save(
          self.directory,
          args=self.save_args_cls(self.pytree, save_args=save_args),
      )
      metadata = self.handler.metadata(self.directory)
      jax.tree.map(lambda m: self.assertEqual(m.dtype, jnp.float32), metadata)

      def check_dtype(x, dtype):
        if utils.is_scalar(x):
          self.assertIsInstance(x, int)
        else:
          self.assertEqual(x.dtype, dtype)

      pytree = jax.tree.map(
          functools.partial(
              utils.to_shape_dtype_struct,
              dtype=jnp.bfloat16,
              scalar_dtype=int,
          ),
          self.pytree,
      )
      restored = self.handler.restore(
          self.directory,
          args=self.restore_args_cls(pytree),
      )
      jax.tree.map(lambda x: check_dtype(x, jnp.bfloat16), restored)

    def test_flax_model(self):
      """Test case."""

      @flax.struct.dataclass
      class Params(flax.struct.PyTreeNode):
        params: Any
        opt_state: Any

      def make_params():
        return Params(
            params=self.numpy_pytree,
            opt_state=(optax.EmptyState(), optax.EmptyState()),
        )

      params = make_params()
      mesh = jax.sharding.Mesh(np.asarray(jax.devices()), ('devices',))
      mesh_axes = jax.sharding.PartitionSpec()
      params = jax.tree.map(
          lambda arr: test_utils.create_sharded_array(arr, mesh, mesh_axes),
          params,
      )
      target = jax.tree.map(utils.to_shape_dtype_struct, params)

      self.handler.save(self.directory, args=self.save_args_cls(params))
      restored = self.handler.restore(
          self.directory, args=self.restore_args_cls(target)
      )
      test_utils.assert_tree_equal(self, params, restored)

    def test_empty_error(self):
      """Test case."""
      with self.assertRaises(ValueError):
        self.handler.save(self.directory, args=self.save_args_cls({}))

    def test_empty_dict_node(self):
      """Test case."""
      item = {'a': {}, 'b': 3}
      self.handler.save(self.directory, args=self.save_args_cls(item))
      restored = self.handler.restore(
          self.directory, args=self.restore_args_cls(item)
      )
      self.assertDictEqual(restored, item)

    def test_empty_none_node(self):
      """Test case."""
      item = {'c': None, 'd': 2}
      self.handler.save(self.directory, args=self.save_args_cls(item))
      restored = self.handler.restore(
          self.directory, args=self.restore_args_cls(item)
      )
      self.assertDictEqual(restored, item)

    def test_none_node_in_restore_args(self):
      """Test case."""
      devices = np.asarray(jax.devices())
      mesh = jax.sharding.Mesh(devices, ('x',))
      mesh_axes = jax.sharding.PartitionSpec(
          'x',
      )
      arr = test_utils.create_sharded_array(np.arange(16), mesh, mesh_axes)
      item = {'b': arr}
      self.handler.save(self.directory, args=self.save_args_cls(item))

      restored = self.handler.restore(
          self.directory,
          args=self.restore_args_cls({'b': None}),
      )
      test_utils.assert_tree_equal(self, restored, {'b': None})

    def test_masked_shape_dtype_struct(self):
      """Test case."""

      def _should_mask(keypath):
        return keypath[0].key == 'a' or (
            keypath[0].key == 'c' and keypath[1].key == 'e'
        )

      def _mask(keypath, x):
        return optax.MaskedNode() if _should_mask(keypath) else x

      def _none(keypath, x):
        return None if _should_mask(keypath) else x

      masked_tree = jax.tree_util.tree_map_with_path(_mask, self.pytree)
      expected = jax.tree_util.tree_map_with_path(_none, self.pytree)

      self.handler.save(self.directory, args=self.save_args_cls(masked_tree))
      self.assertTrue(type_handlers.is_ocdbt_checkpoint(self.directory))

      # Restore it with item which was given before applying masking.
      restored = self.handler.restore(
          self.directory,
          args=self.restore_args_cls(
              jax.tree.map(utils.to_shape_dtype_struct, self.pytree)
          ),
      )
      test_utils.assert_tree_equal(self, expected, restored)

      # Restore it with item after applying masking to it.
      restored = self.handler.restore(
          self.directory,
          args=self.restore_args_cls(
              jax.tree.map(utils.to_shape_dtype_struct, masked_tree)
          ),
      )
      test_utils.assert_tree_equal(self, expected, restored)

      # Restore it without any item.
      restored = self.handler.restore(self.directory)
      test_utils.assert_tree_equal(self, expected, restored)



# Copyright 2024 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for Orbax step storage constructs."""

from typing import Iterator
from unittest import mock
from absl.testing import absltest
from absl.testing import parameterized
from etils import epath
from orbax.checkpoint import test_utils
from orbax.checkpoint._src.metadata import checkpoint
from orbax.checkpoint._src.metadata import step_metadata_serialization
from orbax.checkpoint._src.path import atomicity
from orbax.checkpoint._src.path import step as step_lib


class StandardNameFormatTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self.directory = epath.Path(
        self.create_tempdir(name='step_dir_test').full_path
    )
    self.step_name_prefix = [
        '',
        '0',
        '1',
        '10',
        '11',
        '00',
        '01',
        '010',
        'a',
        'aa',
        'aaa',
        'tmp',
    ]
    self.step_name_prefix_sep = ['', '.', '_']
    self.step_name_step_prefix = ['', 'a', 'aa', 'aaa', 'tmp']
    self.step_name_step_prefix_sep = ['', '.', '_']
    self.step_name_step = [
        '',
        '0',
        '1',
        '10',
        '11',
        '100',
        '101',
        '111',
        '00',
        '01',
        '000',
        '001',
        '010',
        '011',
        'tmp',
    ]

  def generate_step_names(self) -> Iterator[str]:
    for snf in self.step_name_prefix:
      for snps in self.step_name_prefix_sep:
        for snsp in self.step_name_step_prefix:
          for snsps in self.step_name_step_prefix_sep:
            for sns in self.step_name_step:
              yield f'{snf}{snps}{snsp}{snsps}{sns}'

  @parameterized.named_parameters(
      dict(
          testcase_name='step0',
          step_prefix=None,
          step_format_fixed_length=None,
          step=0,
          expected='0',
      ),
      dict(
          testcase_name='step1',
          step_prefix=None,
          step_format_fixed_length=None,
          step=1,
          expected='1',
      ),
      dict(
          testcase_name='padded0',
          step_prefix=None,
          step_format_fixed_length=2,
          step=0,
          expected='00',
      ),
      dict(
          testcase_name='padded1',
          step_prefix=None,
          step_format_fixed_length=2,
          step=1,
          expected='01',
      ),
      dict(
          testcase_name='padded10',
          step_prefix=None,
          step_format_fixed_length=2,
          step=10,
          expected='10',
      ),
      dict(
          testcase_name='prefix_unpadded0',
          step_prefix='aa',
          step_format_fixed_length=None,
          step=0,
          expected='aa_0',
      ),
      dict(
          testcase_name='prefix_unpadded1',
          step_prefix='aa',
          step_format_fixed_length=None,
          step=1,
          expected='aa_1',
      ),
      dict(
          testcase_name='prefix_padded0',
          step_prefix='aa',
          step_format_fixed_length=2,
          step=0,
          expected='aa_00',
      ),
      dict(
          testcase_name='prefix_padded1',
          step_prefix='aa',
          step_format_fixed_length=2,
          step=1,
          expected='aa_01',
      ),
      dict(
          testcase_name='prefix_padded10',
          step_prefix='aa',
          step_format_fixed_length=2,
          step=10,
          expected='aa_10',
      ),
      dict(
          testcase_name='error.step',
          step_prefix=None,
          step_format_fixed_length=None,
          step=5,
          expected=None,
      ),
      dict(
          testcase_name='error.padded',
          step_prefix=None,
          step_format_fixed_length=2,
          step=111111,
          expected=None,
      ),
      dict(
          testcase_name='error.prefix_unpadded',
          step_prefix='aa',
          step_format_fixed_length=None,
          step=5,
          expected=None,
      ),
      dict(
          testcase_name='error.prefix_padded',
          step_prefix='aa',
          step_format_fixed_length=2,
          step=5,
          expected=None,
      ),
  )
  def test_find_step(
      self,
      step_prefix,
      step_format_fixed_length,
      step,
      expected,
  ):
    """Tests for committed steps with `find_step()` and `find_step_path()`."""
    for step_dir in self.generate_step_names():
      (self.directory / step_dir).mkdir(parents=True, exist_ok=True)

    name_format = step_lib.standard_name_format(
        step_prefix=step_prefix,
        step_format_fixed_length=step_format_fixed_length,
    )
    if expected is None:
      with self.assertRaises(ValueError):
        _ = name_format.find_step(self.directory, step)
      with self.subTest('find_step_path_with_include_uncommitted=False'):
        with self.assertRaises(ValueError):
          _ = step_lib.find_step_path(
              self.directory,
              name_format,
              step=step,
              include_uncommitted=False,
          )
    else:
      self.assertEqual(
          self.directory / expected,
          name_format.find_step(self.directory, step).path,
      )
      with self.subTest('uncommitted_gcs_step'):
        with mock.patch.object(
            step_lib, 'is_gcs_path', autospec=True, return_value=True
        ):
          with self.assertRaises(ValueError):
            _ = name_format.find_step(self.directory, step)
      with self.subTest('find_step_path_with_include_uncommitted=False'):
        self.assertEqual(
            self.directory / expected,
            step_lib.find_step_path(
                self.directory,
                name_format,
                step=step,
                include_uncommitted=False,
            ),
        )
      with self.subTest(
          'find_uncommited_gcs_step_path_with_include_uncommitted=False'
      ):
        with mock.patch.object(
            step_lib, 'is_gcs_path', autospec=True, return_value=True
        ):
          with self.assertRaises(ValueError):
            _ = step_lib.find_step_path(
                self.directory,
                name_format,
                step=step,
                include_uncommitted=False,
            )
      with self.subTest('checkpoint_metadata_based_fields'):
        self.assertIsNone(
            name_format.find_step(self.directory, step).init_timestamp_nsecs
        )
        self.assertIsNone(
            name_format.find_step(self.directory, step).commit_timestamp_nsecs
        )

  @parameterized.parameters(True, False)
  def test_find_step_path_with_uncommitted_checkpoint(self, gcs: bool):
    """Tests for `step.find_step_path(include_uncommitted=True)`."""
    name_format = step_lib.standard_name_format(step_prefix='step')
    if gcs:
      uncommitted_step_name = 'step_200'
      for step_name in ['step_1', 'step_100', uncommitted_step_name]:
        step_dir = self.directory / step_name
        step_dir.mkdir(parents=True, exist_ok=True)
        if step_name is not uncommitted_step_name:
          commit_success_file = step_dir / step_lib._COMMIT_SUCCESS_FILE
          commit_success_file.write_text(
              f'Checkpoint commit was successful to {step_dir}'
          )
    else:
      uncommitted_step_name = atomicity._get_tmp_directory(
          step_lib.get_save_directory(
              step=200, directory=self.directory, step_name_format=name_format
          )
      ).name
      for step_name in ['step_1', 'step_100', uncommitted_step_name]:
        (self.directory / step_name).mkdir(parents=True, exist_ok=True)

    if gcs:
      with mock.patch.object(
          step_lib, 'is_gcs_path', autospec=True, return_value=True
      ):
        self.assertEqual(
            self.directory / uncommitted_step_name,
            step_lib.find_step_path(
                self.directory, name_format, step=200, include_uncommitted=True
            ),
        )
        with self.assertRaises(ValueError):
          _ = step_lib.find_step_path(
              self.directory, name_format, step=300, include_uncommitted=True
          )
    else:
      self.assertEqual(
          self.directory / uncommitted_step_name,
          step_lib.find_step_path(
              self.directory, name_format, step=200, include_uncommitted=True
          ),
      )
      with self.assertRaises(ValueError):
        _ = step_lib.find_step_path(
            self.directory, name_format, step=300, include_uncommitted=True
        )

  @parameterized.named_parameters(
      dict(
          testcase_name='step',
          step_prefix=None,
          step_format_fixed_length=None,
          expected=['0', '1', '10', '11', '100', '101', '111'],
      ),
      dict(
          testcase_name='padded',
          step_prefix=None,
          step_format_fixed_length=2,
          expected=['00', '01', '10', '11', '100', '101', '111'],
      ),
      dict(
          testcase_name='prefix_unpadded',
          step_prefix='aa',
          step_format_fixed_length=None,
          expected=[
              'aa_0',
              'aa_1',
              'aa_10',
              'aa_11',
              'aa_100',
              'aa_101',
              'aa_111',
          ],
      ),
      dict(
          testcase_name='prefix_padded',
          step_prefix='aa',
          step_format_fixed_length=2,
          expected=[
              'aa_00',
              'aa_01',
              'aa_10',
              'aa_11',
              'aa_100',
              'aa_101',
              'aa_111',
          ],
      ),
  )
  def test_find_all(
      self,
      step_prefix,
      step_format_fixed_length,
      expected,
  ):
    self.step_name_prefix = ['', '00', '01', '010', 'a', 'aa', 'aaa', 'tmp']

    for step_dir in self.generate_step_names():
      (self.directory / step_dir).mkdir(parents=True, exist_ok=True)

    name_format = step_lib.standard_name_format(
        step_prefix=step_prefix,
        step_format_fixed_length=step_format_fixed_length,
    )
    self.assertSetEqual(
        set(expected),
        set([
            metadata.path.name
            for metadata in name_format.find_all(self.directory)
        ]),
    )
    with self.subTest('find_all_uncommitted_gcs_steps'):
      with mock.patch.object(
          step_lib, 'is_gcs_path', autospec=True, return_value=True
      ):
        self.assertEmpty([
            metadata.path.name
            for metadata in name_format.find_all(self.directory)
        ])


class MetadataTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self.directory = epath.Path(
        self.create_tempdir(name='metadata_test').full_path
    )

  def test_checkpoint_metadata_based_fields(self):
    step_path = self.directory / 'step_1'
    step_path.mkdir(parents=True, exist_ok=True)
    metadata = checkpoint.StepMetadata(
        init_timestamp_nsecs=1,
        commit_timestamp_nsecs=2,
    )
    checkpoint.metadata_store(enable_write=True, blocking_write=True).write(
        file_path=checkpoint.step_metadata_file_path(step_path),
        metadata=step_metadata_serialization.serialize(metadata),
    )

    metadata = step_lib.Metadata(step=1, path=step_path)

    self.assertEqual(metadata.init_timestamp_nsecs, 1)
    self.assertEqual(metadata.commit_timestamp_nsecs, 2)


class UtilsTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self.directory = epath.Path(
        self.create_tempdir(name='checkpointing_test').full_path
    )

  @parameterized.parameters(
      (3, 'dir', None, None, None, 'dir/3'),
      (3, 'dir', 'params', None, None, 'dir/3/params'),
      (3, 'dir', 'params', 'checkpoint', None, 'dir/checkpoint_3/params'),
      (3, 'dir', None, None, 2, 'dir/03'),
      (4000, 'dir', 'params', None, 5, 'dir/04000/params'),
      (555, 'dir', 'params', 'foo', 8, 'dir/foo_00000555/params'),
      (1234567890, 'dir', 'params', 'foo', 12, 'dir/foo_001234567890/params'),
  )
  def test_get_save_directory(
      self,
      step,
      directory,
      name,
      step_prefix,
      step_format_fixed_length,
      result,
  ):
    self.assertEqual(
        step_lib.get_save_directory(
            step,
            directory,
            name=name,
            step_prefix=step_prefix,
            step_format_fixed_length=step_format_fixed_length,
        ),
        epath.Path(result),
    )

  def test_get_save_directory_tmp_dir_override(self):
    self.assertEqual(
        step_lib.get_save_directory(
            42,
            'path/to/my/dir',
            name='params',
            step_prefix='foobar_',
            override_directory='a/different/dir/path',
        ),
        epath.Path('a/different/dir/path/params'),
    )

  @parameterized.parameters((None,), ('checkpoint_',), ('foobar_',))
  def test_is_tmp_checkpoint(self, step_prefix):
    step_dir = step_lib.get_save_directory(
        5, self.directory, step_prefix=step_prefix
    )
    step_dir.mkdir(parents=True)
    self.assertFalse(step_lib.is_tmp_checkpoint(step_dir))
    tmp_step_dir = atomicity._get_tmp_directory(step_dir)
    test_utils.create_tmp_directory(tmp_step_dir, step_dir)
    self.assertTrue(step_lib.is_tmp_checkpoint(tmp_step_dir))

    item_dir = step_lib.get_save_directory(
        10, self.directory, name='params', step_prefix=step_prefix
    )
    item_dir.mkdir(parents=True)
    self.assertFalse(step_lib.is_tmp_checkpoint(item_dir))
    tmp_item_dir = atomicity._get_tmp_directory(item_dir)
    test_utils.create_tmp_directory(tmp_item_dir, item_dir)
    self.assertTrue(step_lib.is_tmp_checkpoint(tmp_item_dir))

  @parameterized.parameters(
      ('0', 0),
      ('0000', 0),
      ('1000', 1000),
      ('checkpoint_0', 0),
      ('checkpoint_0000', 0),
      ('checkpoint_003400', 3400),
      ('foobar_1000', 1000),
      ('0.orbax-checkpoint-tmp-1010101', 0),
      ('0000.orbax-checkpoint-tmp-12323232', 0),
      ('foobar_1.orbax-checkpoint-tmp-12424424', 1),
      ('foobar_000505.orbax-checkpoint-tmp-13124', 505),
      ('checkpoint_16.orbax-checkpoint-tmp-123214324', 16),
  )
  def test_step_from_checkpoint_name(self, name, step):
    self.assertEqual(step_lib.step_from_checkpoint_name(name), step)

  @parameterized.parameters(
      ('abc',),
      ('checkpoint_',),
      ('checkpoint_1010_',),
      ('_191',),
      ('.orbax-checkpoint-tmp-191913',),
      ('0.orbax-checkpoint-tmp-',),
      ('checkpoint_.orbax-checkpoint-tmp-191913',),
  )
  def test_step_from_checkpoint_name_invalid(self, name):
    with self.assertRaises(ValueError):
      step_lib.step_from_checkpoint_name(name)

  def test_checkpoint_steps_paths_nonexistent_directory_fails(self):
    with self.assertRaisesRegex(ValueError, 'does not exist'):
      step_lib.checkpoint_steps_paths('/non/existent/dir')

  def test_checkpoint_steps_paths_returns_finalized_paths(self):
    digit_only_path = epath.Path(self.directory / '2')
    digit_only_path.mkdir()
    prefix_path = epath.Path(self.directory / 'checkpoint_01')
    prefix_path.mkdir()
    epath.Path(self.directory / 'checkpoint').mkdir()
    epath.Path(self.directory / '1000.orbax-checkpoint-tmp-1010101').mkdir()

    self.assertCountEqual(
        step_lib.checkpoint_steps_paths(self.directory),
        [digit_only_path, prefix_path],
    )

  def test_checkpoint_steps_returns_steps_of_finalized_paths(self):
    epath.Path(self.directory / '2').mkdir()
    epath.Path(self.directory / 'checkpoint_01').mkdir()
    epath.Path(self.directory / 'checkpoint').mkdir()
    epath.Path(self.directory / '1000.orbax-checkpoint-tmp-1010101').mkdir()

    self.assertSameElements(
        [1, 2],
        step_lib.checkpoint_steps(self.directory),
    )

  def test_latest_step_metadata(self):
    name_format = step_lib.standard_name_format(step_prefix='step')

    with self.subTest('empty_root'):
      self.assertIsNone(
          step_lib.latest_step_metadata(self.directory, name_format)
      )

    epath.Path(self.directory / '99').mkdir()
    epath.Path(self.directory / 'step_3').mkdir()

    with self.subTest('single_step'):
      metadata = step_lib.latest_step_metadata(self.directory, name_format)
      self.assertIsNotNone(metadata)
      self.assertEqual(metadata.step, 3)

    epath.Path(self.directory / 'step_4').mkdir()

    with self.subTest('step_with_highest_value'):
      metadata = step_lib.latest_step_metadata(self.directory, name_format)
      self.assertIsNotNone(metadata)
      self.assertEqual(metadata.step, 4)

  def test_step_metadata_of_checkpoint_path(self):
    name_format = step_lib.standard_name_format(step_prefix='step')

    with self.subTest('empty_root'):
      with self.assertRaisesRegex(
          ValueError, 'Failed to resolve step metadata of checkpoint path'
      ):
        step_lib.step_metadata_of_checkpoint_path(
            self.directory / '0', name_format
        )

    epath.Path(self.directory / 'step_2').mkdir()

    with self.subTest('good_checkpoint_path_in_root_with_one_step'):
      metadata = step_lib.step_metadata_of_checkpoint_path(
          self.directory / 'step_2', name_format
      )
      self.assertEqual(metadata.step, 2)

    epath.Path(self.directory / 'step_3').mkdir()

    with self.subTest('good_checkpoint_path_in_root_with_many_steps'):
      metadata = step_lib.step_metadata_of_checkpoint_path(
          self.directory / 'step_2', name_format
      )
      self.assertEqual(metadata.step, 2)

    epath.Path(self.directory / '1').mkdir()

    with self.subTest('bad_checkpoint_path'):
      with self.assertRaisesRegex(
          ValueError, 'Failed to resolve step metadata of checkpoint path'
      ):
        step_lib.step_metadata_of_checkpoint_path(
            self.directory / '1', name_format
        )



if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for asyncio_utils module."""

import asyncio
import functools
import timeit

from absl import logging
from absl.testing import absltest
from absl.testing import parameterized
from orbax.checkpoint._src import asyncio_utils


partial = functools.partial


async def one():
  return 1


async def add(a_coro_fn, b_coro_fn):
  a = await a_coro_fn()
  b = await b_coro_fn()
  return a + b


async def nested(a_coro_fn):
  await a_coro_fn()


async def raise_error():
  raise ValueError("test error")


async def with_run_sync(a_coro_fn):
  x = asyncio_utils.run_sync(a_coro_fn())
  y = asyncio_utils.run_sync(a_coro_fn())
  z = asyncio_utils.run_sync(a_coro_fn())
  return f"{x}{y}{z}"


class AsyncioUtilsTest(parameterized.TestCase):

  @parameterized.named_parameters(
      ["basic", one],
      ["nested_1_level", partial(add, one, one)],
      ["nested_2_level", partial(add, partial(add, one, one), one)],
      [
          "nested_3_level",
          partial(add, partial(add, partial(add, one, one), one), one),
      ],
      [
          "nested_4_level",
          partial(
              add,
              partial(add, partial(add, partial(add, one, one), one), one),
              one,
          ),
      ],
      [
          "nested_5_level",
          partial(
              add,
              partial(
                  add,
                  partial(add, partial(add, partial(add, one, one), one), one),
                  one,
              ),
              one,
          ),
      ],
      [
          "nested_6_level",
          partial(
              add,
              partial(
                  add,
                  partial(
                      add,
                      partial(
                          add, partial(add, partial(add, one, one), one), one
                      ),
                      one,
                  ),
                  one,
              ),
              one,
          ),
      ],
  )
  def test_run_sync_basic(self, coro_fn):
    self.assertEqual(
        asyncio.run(coro_fn()),
        asyncio_utils.run_sync(coro_fn()),
    )

  @parameterized.named_parameters(
      ["basic", raise_error],
      ["nested_1_level", partial(nested, raise_error)],
      ["nested_2_level", partial(nested, partial(nested, raise_error))],
      [
          "nested_3_level",
          partial(nested, partial(nested, partial(nested, raise_error))),
      ],
      [
          "nested_4_level",
          partial(
              nested,
              partial(nested, partial(nested, partial(nested, raise_error))),
          ),
      ],
      [
          "nested_5_level",
          partial(
              nested,
              partial(
                  nested,
                  partial(
                      nested, partial(nested, partial(nested, raise_error))
                  ),
              ),
          ),
      ],
      [
          "nested_6_level",
          partial(
              nested,
              partial(
                  nested,
                  partial(
                      nested,
                      partial(
                          nested, partial(nested, partial(nested, raise_error))
                      ),
                  ),
              ),
          ),
      ],
  )
  def test_run_sync_raising_error(self, coro_fn):
    with self.assertRaisesRegex(ValueError, "test error"):
      asyncio.run(coro_fn())
    with self.assertRaisesRegex(ValueError, "test error"):
      asyncio_utils.run_sync(coro_fn())

  @parameterized.named_parameters(
      ["basic", partial(with_run_sync, one)],
      ["nested_1_level", partial(with_run_sync, partial(with_run_sync, one))],
      [
          "nested_2_level",
          partial(
              with_run_sync, partial(with_run_sync, partial(with_run_sync, one))
          ),
      ],
      [
          "nested_3_level",
          partial(
              with_run_sync,
              partial(
                  with_run_sync,
                  partial(with_run_sync, partial(with_run_sync, one)),
              ),
          ),
      ],
      [
          "nested_4_level",
          partial(
              with_run_sync,
              partial(
                  with_run_sync,
                  partial(
                      with_run_sync,
                      partial(with_run_sync, partial(with_run_sync, one)),
                  ),
              ),
          ),
      ],
      [
          "nested_5_level",
          partial(
              with_run_sync,
              partial(
                  with_run_sync,
                  partial(
                      with_run_sync,
                      partial(
                          with_run_sync,
                          partial(with_run_sync, partial(with_run_sync, one)),
                      ),
                  ),
              ),
          ),
      ],
      [
          "nested_6_level",
          partial(
              with_run_sync,
              partial(
                  with_run_sync,
                  partial(
                      with_run_sync,
                      partial(
                          with_run_sync,
                          partial(
                              with_run_sync,
                              partial(
                                  with_run_sync, partial(with_run_sync, one)
                              ),
                          ),
                      ),
                  ),
              ),
          ),
      ],
  )
  def test_run_sync_nested(self, coro_fn):
    self.assertEqual(
        asyncio.run(coro_fn()),
        asyncio_utils.run_sync(coro_fn()),
    )

  def test_run_nested(self):
    async def with_run(a_coro_fn):
      return asyncio.run(a_coro_fn())

    with self.assertRaisesWithLiteralMatch(
        RuntimeError,
        "asyncio.run() cannot be called from a running event loop",
    ):
      asyncio.run(with_run(one))

    with self.assertRaisesWithLiteralMatch(
        RuntimeError,
        "asyncio.run() cannot be called from a running event loop",
    ):
      asyncio_utils.run_sync(with_run(one))

  @absltest.skip("benchmark asyncio.get_running_loop().")
  def test_benchmark_get_running_loop(self):
    def _is_event_loop_present():
      try:
        asyncio.get_running_loop()
        return True
      except RuntimeError:
        return False

    async def _async_is_event_loop_present():
      _is_event_loop_present()

    def _with_event_loop_present():
      asyncio.run(_async_is_event_loop_present())

    nel = timeit.timeit(_is_event_loop_present, number=1000)  # ~0.001s
    wel = timeit.timeit(_with_event_loop_present, number=1000)  # ~0.182s
    logging.info("time: no_event_loop=%s, with_event_loop=%s", nel, wel)

  @absltest.skip("benchmark asyncio_utils.run_sync.")
  def test_benchmark_run_sync(self):
    async def _test():
      return partial(
          add,
          partial(
              add,
              partial(
                  add,
                  partial(add, partial(add, partial(add, one, one), one), one),
                  one,
              ),
              one,
          ),
          one,
      )()

    number = 10000
    # First run with enable_nest_asyncio=False, because nest_asyncio.apply
    # patches asyncio globally in a runtime. There is no way to unpatch it.
    run_time = timeit.timeit(
        lambda: asyncio_utils.run_sync(_test(), enable_nest_asyncio=False),
        number=number,
    )  # ~1.604s
    run_sync_time = timeit.timeit(
        lambda: asyncio_utils.run_sync(_test(), enable_nest_asyncio=True),
        number=number,
    )  # ~1.5503s
    logging.info(
        "time: run_sync_time=%s, run_time=%s, ratio=%s",
        run_sync_time,
        run_time,
        run_sync_time / run_time,
    )


if __name__ == "__main__":
  absltest.main()



# Copyright 2024 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Registry for checkpoint handlers (:py:class:`CheckpointHandler`).

A checkpoint handler registry is used to register a
:py:class:`CheckpointHandler` for specific a checkpointable item name and a
:py:class:`CheckpointArgs` type. When saving or restoring a checkpoint, the
registry can be used to find the appropriate handler for the item and its args.
The default implementation of the registry is
:py:class:`DefaultCheckpointHandlerRegistry`. See the docstring for the default
implementation for an example how to use the registry.
"""

from typing import MutableMapping, Optional, Protocol, Type, Union

from orbax.checkpoint import checkpoint_args
from orbax.checkpoint._src.handlers import checkpoint_handler

CheckpointArgs = checkpoint_args.CheckpointArgs
CheckpointHandler = checkpoint_handler.CheckpointHandler
HandlerRegistryMapping = MutableMapping[
    tuple[Optional[str], type[CheckpointArgs]], CheckpointHandler
]


class CheckpointHandlerRegistry(Protocol):
  """Protocol for `CheckpointHandlerRegistry`.

  A checkpoint handler registry is used to register
  :py:class:`CheckpointHandler`s for specific item names and
  :py:class:`CheckpointArgs` types. When saving or restoring a checkpoint, the
  registry can be used to find the appropriate handler for the item and its
  args. See :py:class:`DefaultCheckpointHandlerRegistry` for the default
  implementation.
  """

  def add(
      self,
      item: Optional[str],
      args: Union[Type[CheckpointArgs], CheckpointArgs],
      handler: Optional[
          Union[CheckpointHandler, Type[CheckpointHandler]]
      ] = None,
      **kwargs,
  ) -> None:
    """Adds an entry to the registry."""
    ...

  def get(
      self,
      item: Optional[str],
      args: Union[Type[CheckpointArgs], CheckpointArgs],
  ) -> CheckpointHandler:
    """Gets an entry from the registry."""
    ...

  def has(
      self,
      item: Optional[str],
      args: Union[Type[CheckpointArgs], CheckpointArgs],
  ) -> bool:
    """Checks if an entry exists in the registry."""
    ...

  def get_all_entries(
      self,
  ) -> HandlerRegistryMapping:
    ...


class AlreadyExistsError(ValueError):
  """Raised when an entry already exists in the registry."""


class NoEntryError(KeyError):
  """Raised when no entry exists in the registry."""


def _get_args_type(
    args: Union[Type[CheckpointArgs], CheckpointArgs],
) -> type[CheckpointArgs]:
  if isinstance(args, type):
    return args
  else:
    return type(args)


class DefaultCheckpointHandlerRegistry(CheckpointHandlerRegistry):
  """Default implementation of :py:class:`CheckpointHandlerRegistry`.

  Usage::
    from orbax.checkpoint import handlers

    # Create a handler registry.
    registry = handlers.DefaultCheckpointHandlerRegistry()

    # Add custom save and restore args for a checkpointable item.
    registry.add(
        'item',
        FooCustomCheckpointSaveArgs,
        FooCustomCheckpointHandler,
    )
    registry.add(
        'item',
        FooCustomCheckpoinRestoreArgs,
        FooCustomCheckpointHandler,
    )

    # Retrieve the handler for the item and args.
    handler = registry.get('item', FooCustomCheckpointSaveArgs)
    assert isinstance(handler, FooCustomCheckpointHandler)


    # You can also register a handler for a general args type.
    registry.add(
        None,
        BarCustomCheckpointArgs,
        BarCustomCheckpointHandler,
    )

    # Retrieve the handler for the general args type (for an item which has
    # not been registered).
    handler = registry.get(None, BarCustomCheckpointSaveArgs)
    assert isinstance(handler, BarCustomCheckpointHandler)

    # Associate a `CheckpointArgs` (and `CheckpointHandler`) with a given item
    # name, provided that the `CheckpointHandler` is globally registered.
    registry.add('state', ocp.args.StandardSave)
    registry.add('state', ocp.args.StandardRestore)
  """

  def __init__(
      self, other_registry: Optional[CheckpointHandlerRegistry] = None
  ):
    self._registry: HandlerRegistryMapping = {}

    # Initialize the registry with entries from other registry.
    if other_registry:
      for (
          item,
          args_type,
      ), handler in other_registry.get_all_entries().items():
        self.add(item, args_type, handler)

  def add(
      self,
      item: Optional[str],
      args: Union[Type[CheckpointArgs], CheckpointArgs],
      handler: Optional[
          Union[CheckpointHandler, Type[CheckpointHandler]]
      ] = None,
  ):
    """Adds an entry to the registry.

    Args:
      item: The item name. If None, the entry will be added as a general `arg`
        entry.
      args: The args type to be added to registry. If a concerete type is
        provided, the type will be added to the registry.
      handler: The handler. If a type is provided, an instance of the type will
        be added to the registry. If None, tries to find a globally registered
        handler for the given args type.

    Raises:
      AlreadyExistsError: If an entry for the given item and args type already
        exists in the registry.
      ValueError: If no globally registered handler can be found for the given
        args type and handler is provided as None or if the handler is globally
        registered but not default-constructible.
    """
    args_type = _get_args_type(args)

    if self.has(item, args_type):
      raise AlreadyExistsError(
          f'Entry for item={item} and args_type={args_type} already'
          ' exists in the registry.'
      )
    if handler is None:
      try:
        handler_to_register = checkpoint_args.get_registered_handler_cls(args)
      except ValueError as e:
        raise ValueError(
            'Provided a `None` `CheckpointHandler` when registering'
            f' item={item} and args_type={args_type}. Unable to find a globally'
            ' registered handler for the given args type. Provide a `handler`'
            ' or ensure the args type corresponds to a globally registered'
            ' handler.'
        ) from e
    else:
      handler_to_register = handler

    if isinstance(handler_to_register, type):
      try:
        handler_instance = handler_to_register()
      except TypeError as e:
        raise ValueError(
            'Provided a `None` `CheckpointHandler` when registering'
            f' item={item} and args_type={args_type}. The globally-registered'
            ' handler for this args type is not default-constructible. Please'
            ' a `handler` when adding this args type to the registry.'
        ) from e
    else:
      handler_instance = handler_to_register

    self._registry[(item, args_type)] = handler_instance

  def get(
      self,
      item: Optional[str],
      args: Union[Type[CheckpointArgs], CheckpointArgs],
  ) -> CheckpointHandler:
    """Returns the handler for the given item and args type.

    Args:
      item: The item name. If None, the entry will be added as a general type
        `args` entry.
      args: The args type to get the handler for. If a concerete type is
        provided, the type will be used to get the handler.

    Raises:
      NoEntryError: If no entry for the given item and args exists in the
        registry.
    """
    args_type = _get_args_type(args)

    if self.has(item, args_type):
      return self._registry[(item, args_type)]

    raise NoEntryError(
        f'No entry for item={item} and args_ty={args_type} in the registry.'
    )

  def has(
      self,
      item: Optional[str],
      args: Union[Type[CheckpointArgs], CheckpointArgs],
  ) -> bool:
    """Returns whether an entry for the given item and args type exists in the registry.

    Args:
      item: The item name or None.
      args: The args type to check for. If a concrete type is provided, the type
        will be used to check for the entry.
    """
    args_type = _get_args_type(args)
    return (
        item,
        args_type,
    ) in self._registry

  def get_all_entries(
      self,
  ) -> HandlerRegistryMapping:
    """Returns all entries in the registry."""
    return self._registry

  def __repr__(self):
    return f'DefaultCheckpointHandlerRegistry({self._registry})'

  def __str__(self):
    return f'DefaultCheckpointHandlerRegistry({self._registry})'



# Copyright 2024 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Public common types to work with pytrees."""

from typing import Any, TypeVar, Union

from jax import tree_util as jtu


PyTree = Any

# This trick doesn't allow us to achieve full type checking but we at least can
# use concrete types to document things like `PyTreeOf[jax.Array]`.
T = TypeVar('T')
PyTreeOf = Union[PyTree, T]

PyTreeKey = Union[
    jtu.SequenceKey, jtu.DictKey, jtu.GetAttrKey, jtu.FlattenedIndexKey
]
PyTreePath = tuple[PyTreeKey, ...]



# Copyright 2024 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Leaf of a JSON metadata tree."""

from __future__ import annotations

import dataclasses
from typing import Any, Dict

from orbax.checkpoint._src.metadata import empty_values
from orbax.checkpoint._src.serialization import types


_VALUE_TYPE = 'value_type'
_SKIP_DESERIALIZE = 'skip_deserialize'


@dataclasses.dataclass
class ValueMetadataEntry:
  """Represents metadata for a leaf in a tree.

  WARNING: Do not rename this class, as it is saved by its name in the metadata
  storage.

  IMPORTANT: Please make sure that changes in attributes are backwards
  compatible with existing mmetadata in storage.
  """

  value_type: str
  skip_deserialize: bool = False

  def to_json(self) -> Dict[str, Any]:
    return {
        _VALUE_TYPE: self.value_type,
        _SKIP_DESERIALIZE: self.skip_deserialize,
    }

  @classmethod
  def from_json(cls, json_dict: Dict[str, Any]) -> ValueMetadataEntry:
    return ValueMetadataEntry(
        value_type=json_dict[_VALUE_TYPE],
        skip_deserialize=json_dict[_SKIP_DESERIALIZE],
    )

  @classmethod
  def build(
      cls,
      info: types.ParamInfo,
      save_arg: types.SaveArgs,
  ) -> ValueMetadataEntry:
    """Builds a ValueMetadataEntry."""
    del save_arg
    if info.value_typestr is None:
      raise AssertionError(
          'Must set `value_typestr` in `ParamInfo` when saving.'
      )
    skip_deserialize = empty_values.is_empty_typestr(info.value_typestr)
    return ValueMetadataEntry(
        value_type=info.value_typestr, skip_deserialize=skip_deserialize
    )



# Copyright 2024 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for transform_utils."""

from typing import List, Mapping

from absl.testing import absltest
# TODO(b/275613424): Eliminate flax dependency in Orbax test suite.
import flax
import flax.linen as nn
from flax.training.train_state import TrainState
import jax
from jax import numpy as jnp
import numpy as np
import optax
from orbax.checkpoint import test_utils
from orbax.checkpoint import transform_utils
from orbax.checkpoint._src.tree import utils as tree_utils


Transform = transform_utils.Transform
apply_transformations = transform_utils.apply_transformations


def empty_pytree(tree):
  return jax.tree.map(lambda x: object(), tree)


class EmptyNode:
  pass


# Not in common util because we need to eliminate OSS dependency on flax.
def init_flax_model(model):
  params = model.init(jax.random.PRNGKey(0), jnp.ones([8, 8]))
  tx = optax.adamw(learning_rate=0.001)
  state = TrainState.create(apply_fn=model.apply, params=params, tx=tx)
  return jax.tree.map(np.asarray, state)


class TransformUtilsTest(absltest.TestCase):

  def setUp(self):
    super().setUp()
    self.original = {
        'a': 0,
        'b': 1,
        'c': {
            'a': 2,
            'e': 3,
        }
    }

  def test_empty(self):
    self.assertDictEqual({}, apply_transformations({}, {}, {}))
    self.assertDictEqual({}, apply_transformations({'a': 0, 'b': 1}, {}, {}))
    self.assertDictEqual({'a': 1}, apply_transformations({}, {}, {'a': 1}))

  def test_no_transform(self):
    transforms = jax.tree.map(lambda _: Transform(), self.original)
    self.assertDictEqual(
        self.original,
        apply_transformations(self.original, transforms,
                              empty_pytree(self.original)))

  def test_transform_missing_in_original(self):
    original = {'a': 1}
    transforms = {'b': Transform()}
    with self.assertRaises(ValueError):
      apply_transformations(original, transforms, {'b': ...})

  def test_rename(self):
    transforms = {
        'a1': Transform(original_key='a'),  # originally named "a"
        'c': {
            'a': Transform(),  # unchanged
        },
        # moved from being inside "c"
        'e1': Transform(original_key='c/e'),
        'f': Transform(use_fallback=True),  # newly added
        # note: dropped "b"
        # copied c/a and moved up
        'ca1': Transform(original_key='c/a'),
        'ca2': Transform(original_key='c/a'),
    }
    fallback = {
        'a1': ...,
        'c': {
            'a': ...,
        },
        'e1': ...,
        'f': None,
        'g': 100,  # newly added
        'ca1': ...,
        'ca2': ...,
    }
    expected = {
        'a1': 0,
        'c': {
            'a': 2,
        },
        'e1': 3,
        'f': None,
        'g': 100,  # newly added
        'ca1': 2,
        'ca2': 2,
    }
    self.assertDictEqual(
        expected, apply_transformations(self.original, transforms, fallback))

  def test_partial_transformation(self):
    transforms = {
        'a1': Transform(original_key='a'),  # originally named "a"
        # implicit "a" also gets preserved
        # implicit copy over "c" subtree
        # moved from being inside "c"
        'e1': Transform(original_key='c/e'),
        'e2': Transform(original_key='c/e'),
        # implicit add "f" and "g"
    }
    fallback = {
        'a': ...,
        'a1': ...,
        'c': {
            'a': ...,
            'e': ...,
        },
        'e1': ...,
        'e2': ...,
        'f': None,
        'g': 2,
    }
    expected = {
        'a': 0,
        'a1': 0,
        'c': {
            'a': 2,
            'e': 3,
        },
        'e1': 3,
        'e2': 3,
        'f': None,
        'g': 2,
    }
    self.assertDictEqual(
        expected, apply_transformations(self.original, transforms, fallback))

  def test_default_new(self):
    transforms = {
        'a': Transform(use_fallback=True),  # use value from original
        # implicit drop "b"
        # implicit retain "c/a", "c/a"
        'b1': Transform(original_key='b'),
        # implicit add "f" and "g"
    }
    new = {
        'a': ...,
        'c': {
            'a': 10,
            'e': 11,
        },
        'b1': ...,
        'f': None,
        'g': 2,
    }
    expected = {
        'a': 0,
        'c': {
            'a': 10,
            'e': 11,
        },
        'b1': 1,
        'f': None,
        'g': 2,
    }
    self.assertDictEqual(
        expected,
        apply_transformations(
            self.original, transforms, new, default_to_original=False))

  def test_missing_key_default(self):
    transforms = {'f': Transform(use_fallback=True)}
    new = {
        'a': 2,
        'b': 3,
        'c': {
            'a': 7,
            'e': 8,
        },
        'f': 20,
    }
    with self.assertRaises(ValueError):
      apply_transformations(
          self.original, transforms, new, default_to_original=False)

    expected = {'a': 0, 'b': 1, 'c': {'a': 2, 'e': 3}, 'f': 20}
    self.assertDictEqual(
        expected,
        apply_transformations(
            self.original, transforms, new, default_to_original=True))

  def test_regex(self):
    original = {
        'a1': 1,
        'c': {
            'a2': {
                'd': 2,
            },
            'a3': 3,
        },
        'd': {
            'e1': 2,
            'e2': 4,
        },
        'f': {
            'e3': 6,
            'e4': {  # note doesn't get matched
                'a': 1
            },
        },
    }
    transforms = {
        r'(.*)x(\d.*)':
            Transform(original_key=r'\1a\2'),
        r'(.*)y(\d)':
            Transform(original_key=r'\1e\2', value_fn=lambda val: val * 2),
    }
    expected = {
        'x1': 1,
        'c': {
            'x2': {
                'd': 2,
            },
            'x3': 3,
        },
        'd': {
            'y1': 4,
            'y2': 8,
        },
        'f': {
            'y3': 12,
        },
    }
    self.assertDictEqual(
        expected,
        apply_transformations(original, transforms, empty_pytree(expected)))

  def test_partial_restore(self):
    transforms = {
        'a': Transform(),
        'c': {
            'e': Transform(),
        },
    }
    expected = {
        'a': 0,
        'c': {
            'e': 3,
        },
    }
    self.assertDictEqual(
        expected,
        apply_transformations(self.original, transforms,
                              empty_pytree(expected)))

  def test_function(self):
    transforms = {
        'a': Transform(multi_value_fn=lambda _, kv: kv['a'] * 2 + 20),
        # dropped b
        'c': {
            # added together two keys, leaving one remaining
            'a': Transform(
                multi_value_fn=lambda _, kv: kv['c']['a'] + kv['c']['e']
            ),
        },
        # many to many transformation: input two keys -> output two new keys
        'w': Transform(multi_value_fn=lambda _, kv: kv['a'] + kv['b']),
        'x': Transform(multi_value_fn=lambda _, kv: kv['a'] + kv['b'] * 2),
        # copied a single key into multiple
        'y': Transform(multi_value_fn=lambda _, kv: kv['a']),
        'z': Transform(multi_value_fn=lambda _, kv: kv['a']),
    }
    expected = {
        'a': 20,
        'c': {
            'a': 5,
        },
        'w': 1,
        'x': 2,
        'y': 0,
        'z': 0,
    }
    self.assertDictEqual(
        expected,
        apply_transformations(self.original, transforms,
                              empty_pytree(expected)))

  def test_non_dict_tree(self):

    @flax.struct.dataclass
    class SubTree(flax.struct.PyTreeNode):
      x: Mapping[str, int]
      y: List[int]

    @flax.struct.dataclass
    class Tree(flax.struct.PyTreeNode):
      a: int
      b: np.ndarray
      c: SubTree

    tree = Tree(
        a=10, b=np.arange(3), c=SubTree(x={
            'i': 0,
            'j': 1
        }, y=[4, 5, 6]))

    @flax.struct.dataclass
    class NewTree(flax.struct.PyTreeNode):
      a1: int  # a
      b: np.ndarray  # times 2
      c: SubTree  # same
      d: float  # new
      e: int  # from a.y
      f: List[int]  # from a.y

    transforms = NewTree(
        a1=Transform(original_key='a'),
        b=Transform(multi_value_fn=lambda _, t: t.b * 2),
        c=jax.tree.map(lambda _: Transform(), tree.c),
        d=Transform(use_fallback=True),
        e=Transform(multi_value_fn=lambda _, t: t.c.y[0]),
        f=[
            Transform(multi_value_fn=lambda _, t: t.c.y[1]),
            Transform(multi_value_fn=lambda _, t: t.c.y[2]),
        ],
    )
    fallback_tree = NewTree(
        a1=EmptyNode(),
        b=EmptyNode(),
        c=SubTree(
            x={'i': EmptyNode(), 'j': EmptyNode()},
            y=[EmptyNode(), EmptyNode(), EmptyNode()],
        ),
        d=7,
        e=EmptyNode(),
        f=[EmptyNode(), EmptyNode()],
    )
    expected_tree = NewTree(
        a1=10,
        b=np.arange(3) * 2,
        c=SubTree(x={
            'i': 0,
            'j': 1
        }, y=[4, 5, 6]),
        d=7,
        e=4,
        f=[5, 6])

    def assert_equal(a, b):
      if isinstance(a, np.ndarray):
        np.testing.assert_equal(a, b)
      elif isinstance(a, list):
        self.assertListEqual(a, b)
      elif isinstance(a, dict):
        self.assertDictEqual(a, b)
      else:
        self.assertEqual(a, b)

    jax.tree.map(
        assert_equal, expected_tree,
        apply_transformations(tree, transforms, fallback_tree))

  def test_flax_train_state(self):

    class SmallModel(nn.Module):

      @nn.compact
      def __call__(self, x):
        x = x.reshape((x.shape[0], -1))
        x = nn.Dense(features=8)(x)
        x = nn.sigmoid(x)
        x = nn.Dense(features=8)(x)
        return x

    old_state = init_flax_model(SmallModel())

    class LargeModel(nn.Module):

      @nn.compact
      def __call__(self, x):
        x = x.reshape((x.shape[0], -1))  # flatten
        x = nn.Dense(features=16)(x)
        x = nn.sigmoid(x)
        x = nn.Dense(features=8)(x)
        x = nn.sigmoid(x)
        x = nn.Dense(features=8)(x)
        x = nn.sigmoid(x)
        x = nn.Dense(features=4)(x)
        return x

    new_state = init_flax_model(LargeModel())

    transformations = {
        # LargeModel layer 0 is a newly inserted layer, thus use_fallback=True.
        r'(.*)Dense_0(.*)': Transform(use_fallback=True),
        # SmallModel layer 0 maps to LargeModel layer 1
        r'(.*)Dense_1(.*)': Transform(original_key=r'\1Dense_0\2'),
        # SmallModel layer 1 maps to LargeModel layer 2
        r'(.*)Dense_2(.*)': Transform(original_key=r'\1Dense_1\2')
    }  # Note: LargeModel layer 3 is newly added.
    restored_state = apply_transformations(old_state, transformations,
                                           new_state)

    # Construct expected tree
    old_flat_dict = tree_utils.to_flat_dict(old_state, sep='/')
    new_flat_dict = tree_utils.to_flat_dict(new_state, sep='/')
    expected_flat_dict = {}
    for k, v in new_flat_dict.items():
      if 'Dense_1' in k:
        expected_flat_dict[k] = old_flat_dict[k.replace('Dense_1', 'Dense_0')]
      elif 'Dense_2' in k:
        expected_flat_dict[k] = old_flat_dict[k.replace('Dense_2', 'Dense_1')]
      elif 'Dense_' in k:  # layers in new, but not old.
        expected_flat_dict[k] = v
      else:  # extra keys in both, expected is the old value
        expected_flat_dict[k] = old_flat_dict[k]

    expected_state = tree_utils.from_flat_dict(
        expected_flat_dict, target=new_state, sep='/'
    )
    test_utils.assert_tree_equal(self, expected_state, restored_state)

  def test_flax_train_state_default_new(self):

    class Model(nn.Module):

      @nn.compact
      def __call__(self, x):
        x = x.reshape((x.shape[0], -1))  # flatten
        x = nn.Dense(features=16)(x)
        x = nn.sigmoid(x)
        x = nn.Dense(features=8)(x)
        x = nn.sigmoid(x)
        x = nn.Dense(features=8)(x)
        x = nn.sigmoid(x)
        x = nn.Dense(features=4)(x)
        return x

    old_state = init_flax_model(Model())
    new_state = init_flax_model(Model())

    transformations = {
        # values default to new_state, use_fallback=True instructs the Transform
        # to fall back on old_state for this key.
        r'(.*)Dense_1(.*)': Transform(use_fallback=True),
    }
    restored_state = apply_transformations(
        old_state, transformations, new_state, default_to_original=False)

    # Construct expected tree
    old_flat_dict = tree_utils.to_flat_dict(old_state, sep='/')
    new_flat_dict = tree_utils.to_flat_dict(new_state, sep='/')
    expected_flat_dict = {}
    for k, v in new_flat_dict.items():
      if 'Dense_1' in k:
        expected_flat_dict[k] = old_flat_dict[k]
      else:
        expected_flat_dict[k] = v

    expected_state = tree_utils.from_flat_dict(
        expected_flat_dict, target=new_state, sep='/'
    )
    test_utils.assert_tree_equal(self, expected_state, restored_state)

  def test_merge_trees(self):
    one = {
        'a': 1,
        'b': {
            'c': 2,
            'd': 3,
        },
    }
    two = {
        'a': 2,
        'b': {
            'c': 4,
            'e': 6,
        },
    }
    three = {}
    four = {
        'f': 7,
        'g': 8,
    }
    expected = {
        'a': 2,
        'b': {
            'c': 4,
            'd': 3,
            'e': 6,
        },
        'f': 7,
        'g': 8,
    }
    self.assertDictEqual(
        expected, transform_utils.merge_trees(one, two, three, four)
    )

  def test_merge_trees_target(self):
    one = {
        'a': 1,
        'b': {
            'c': 2,
            'd': 3,
        },
    }
    two = {
        'a': 2,
        'b': {
            'c': 4,
            'e': 6,
        },
    }
    three = {}
    four = {
        'f': 7,
        'g': 8,
    }

    @flax.struct.dataclass
    class Expected(flax.struct.PyTreeNode):
      a: int
      b: Mapping[str, int]
      f: int
      g: int

      def __eq__(self, other):
        return (
            self.a == other.a
            and self.b == other.b
            and self.f == other.f
            and self.g == other.g
        )

    expected = Expected(
        a=2,
        b={
            'c': 4,
            'd': 3,
            'e': 6,
        },
        f=7,
        g=8,
    )

    self.assertEqual(
        expected,
        transform_utils.merge_trees(
            one,
            two,
            three,
            four,
            target=jax.tree.map(lambda x: 0, expected),
        ),
    )

  def test_merge_trees_empty(self):
    self.assertDictEqual({}, transform_utils.merge_trees({}, {}))


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import dataclasses
from typing import Optional, Type, Union

from absl.testing import absltest
from absl.testing import parameterized
from etils import epath
from orbax.checkpoint import checkpoint_args
from orbax.checkpoint._src.handlers import checkpoint_handler
from orbax.checkpoint._src.handlers import handler_registration
from orbax.checkpoint._src.handlers import standard_checkpoint_handler


CheckpointHandler = checkpoint_handler.CheckpointHandler
DefaultCheckpointHandlerRegistry = (
    handler_registration.DefaultCheckpointHandlerRegistry
)
AlreadyExistsError = handler_registration.AlreadyExistsError
NoEntryError = handler_registration.NoEntryError


class _TestCheckpointHandler(CheckpointHandler):
  """No-op checkpoint handler for testing."""

  def save(self, directory: epath.Path, *args, **kwargs) -> None:
    del directory, args, kwargs

  def restore(self, directory: epath.Path, *args, **kwargs) -> None:
    del directory, args, kwargs


@dataclasses.dataclass
class _TestArgs(checkpoint_args.CheckpointArgs):
  """No-op checkpoint args for testing."""
  ...


class _TestHandlerWithRequiredInit(_TestCheckpointHandler):

  def __init__(self, foo: str):
    super().__init__()
    self._foo = foo


@checkpoint_args.register_with_handler(
    _TestHandlerWithRequiredInit, for_save=True, for_restore=True
)
@dataclasses.dataclass
class _TestArgsWithRequiredInit(checkpoint_args.CheckpointArgs):
  """No-op checkpoint args for testing."""

  ...


class HandlerRegistryTest(parameterized.TestCase):

  @parameterized.product(
      handler=(_TestCheckpointHandler, _TestCheckpointHandler()),
      item=(None, 'item'),
  )
  def test_add_and_get_entry(
      self,
      handler: Union[CheckpointHandler, Type[CheckpointHandler]],
      item: Optional[str],
  ):
    args_type = _TestArgs
    registry = DefaultCheckpointHandlerRegistry()

    registry.add(
        item,
        args_type,
        handler,
    )

    # Check that the entry is added to the registry.
    self.assertTrue(registry.has(item, args_type))
    # Check that the handler is returned and that it is initialized as an
    # object.
    self.assertIsInstance(
        registry.get(item, args_type),
        _TestCheckpointHandler,
    )

  def test_add_entry_with_existing_item_and_args_type_raises_error(self):
    item = 'item'
    args_type = _TestArgs
    registry = DefaultCheckpointHandlerRegistry()

    registry.add(item, args_type, _TestCheckpointHandler)

    with self.assertRaisesRegex(
        AlreadyExistsError, r'already exists in the registry'
    ):
      registry.add(item, args_type, _TestCheckpointHandler)

  def test_get_all_entries(self):
    item1 = 'item1'
    item2 = 'item2'
    args_type = _TestArgs
    handler = _TestCheckpointHandler
    registry = DefaultCheckpointHandlerRegistry()

    registry.add(item1, args_type, handler)
    registry.add(item2, args_type, handler)

    entries = registry.get_all_entries()
    self.assertLen(entries, 2)
    self.assertIsInstance(
        entries[(item1, args_type)],
        handler,
    )
    self.assertIsInstance(
        entries[(item2, args_type)],
        handler,
    )

  def test_instantiate_registry_from_another_registry(self):
    item1 = 'item1'
    item2 = 'item2'
    args_type = _TestArgs
    handler = _TestCheckpointHandler

    registry1 = DefaultCheckpointHandlerRegistry()
    registry1.add(item1, args_type, handler)
    registry2 = DefaultCheckpointHandlerRegistry(registry1)
    registry2.add(item2, args_type, handler)

    entries = registry2.get_all_entries()
    self.assertLen(entries, 2)
    self.assertIsInstance(
        entries[(item1, args_type)],
        handler,
    )
    self.assertIsInstance(
        entries[(item2, args_type)],
        handler,
    )

  @parameterized.product(
      item=(None, 'item'),
  )
  def test_raise_error_when_no_entry_found(self, item: Optional[str]):
    registry = DefaultCheckpointHandlerRegistry()

    with self.assertRaisesRegex(
        NoEntryError,
        r'No entry for item=.* and args_ty=.* in the registry',
    ):
      registry.get(item, _TestArgs)

  def test_multiple_handlers_for_same_item(self):
    item = 'item'

    class _TestArgsA(checkpoint_args.CheckpointArgs):
      pass

    class _TestArgsB(checkpoint_args.CheckpointArgs):
      pass

    registry = DefaultCheckpointHandlerRegistry()
    registry.add(item, _TestArgsA, _TestCheckpointHandler)
    registry.add(item, _TestArgsB, _TestCheckpointHandler)

    self.assertIsInstance(
        registry.get(item, _TestArgsA),
        _TestCheckpointHandler,
    )
    self.assertIsInstance(
        registry.get(item, _TestArgsB),
        _TestCheckpointHandler,
    )

  def test_concrete_args_registered_as_type(self):
    item = 'item'
    registry = DefaultCheckpointHandlerRegistry()
    registry.add(item, _TestArgs(), _TestCheckpointHandler)

    # Check that handler can be retrieved using both the concrete type and the
    # type itself.
    self.assertIsInstance(
        registry.get(item, _TestArgs()),
        _TestCheckpointHandler,
    )
    self.assertIsInstance(
        registry.get(item, _TestArgs),
        _TestCheckpointHandler,
    )
    self.assertTrue(registry.has(item, _TestArgs))
    self.assertTrue(registry.has(item, _TestArgs()))

    with self.assertRaisesRegex(
        AlreadyExistsError,
        rf'Entry for item={item} and args_type={_TestArgs}',
    ):
      registry.add(item, _TestArgs, _TestCheckpointHandler)

  @parameterized.product(
      item=(None, 'item'),
  )
  def test_register_with_default_handler(self, item: Optional[str]):
    registry = DefaultCheckpointHandlerRegistry()
    with self.assertRaisesRegex(
        ValueError, 'Provided a `None` `CheckpointHandler`'
    ):
      registry.add(item, _TestArgs)

    registry.add(item, standard_checkpoint_handler.StandardSaveArgs)
    self.assertIsInstance(
        registry.get(item, standard_checkpoint_handler.StandardSaveArgs),
        standard_checkpoint_handler.StandardCheckpointHandler,
    )

  @parameterized.product(
      item=(None, 'item'),
  )
  def test_register_with_default_handler_required_init(
      self, item: Optional[str]
  ):
    registry = DefaultCheckpointHandlerRegistry()
    with self.assertRaisesRegex(ValueError, 'not default-constructible'):
      registry.add(item, _TestArgsWithRequiredInit)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Orbax step storage entities."""

import abc
import concurrent
import dataclasses
import datetime
import functools
import re
import time
from typing import Callable, Generic, Iterator, List, Optional, Protocol, Sequence, Set, TypeVar

from absl import logging
from etils import epath
import jax
import numpy as np
from orbax.checkpoint._src.metadata import checkpoint
from orbax.checkpoint._src.metadata import step_metadata_serialization
from orbax.checkpoint._src.multihost import multihost


_GCS_PATH_PREFIX = ('gs://',)
_COMMIT_SUCCESS_FILE = 'commit_success.txt'
TMP_DIR_SUFFIX = '.orbax-checkpoint-tmp-'
# prefix_1000.orbax-checkpoint-tmp-1010101
# OR
# 1000.orbax-checkpoint-tmp-1010101
TMP_DIR_STEP_PATTERN = r'.*?_*?(\d+)\.orbax-checkpoint-tmp-\d+'
_LAST_CHECKPOINT_WRITE_TIME = time.time()

# This file mode gives full permissions to OWNER, GROUP and OTHER.
WORLD_READABLE_MODE = 0o777

MetadataT = TypeVar('MetadataT', bound='Metadata')


@dataclasses.dataclass(frozen=True)
class Metadata:
  """Metadata of a step.

  Attributes:
    step: step number of the checkpoint.
    path: path to the checkpoint.
  """

  step: int
  path: epath.Path

  @functools.cached_property
  def _checkpoint_metadata(self) -> Optional[checkpoint.StepMetadata]:
    """Returns checkpoint metadata of this step."""
    metadata_dict = checkpoint.metadata_store(enable_write=False).read(
        file_path=checkpoint.step_metadata_file_path(self.path)
    )
    if metadata_dict is None:
      return None
    return step_metadata_serialization.deserialize(metadata_dict)

  @property
  def init_timestamp_nsecs(self) -> Optional[int]:
    """Returns init timestamp of uncommitted checkpoint of this step.

    It is specified as nano seconds since epoch.
    """
    metadata = self._checkpoint_metadata
    if metadata is None:
      return None
    return metadata.init_timestamp_nsecs

  @property
  def commit_timestamp_nsecs(self) -> Optional[int]:
    """Returns commit timestamp of the checkpoint of this step.

    It is specified as nano seconds since epoch.
    """
    metadata = self._checkpoint_metadata
    if metadata is None:
      return None
    return metadata.commit_timestamp_nsecs

  @property
  def commit_timestamp(self) -> datetime.datetime:
    """Returns commit timestamp of the checkpoint of this step.

    It is specified as datetime in UTC timezone.
    """
    commit_timestamp_nsecs = self.commit_timestamp_nsecs
    if commit_timestamp_nsecs is not None:
      timestamp_sec = commit_timestamp_nsecs / 1e9
    else:
      timestamp_sec = self.path.stat().mtime
    return datetime.datetime.fromtimestamp(
        timestamp_sec, tz=datetime.timezone.utc
    )


class NameFormat(Protocol, Generic[MetadataT]):
  """Protocol responsible for naming and querying steps."""

  def build_name(self, step: int) -> str:
    """Returns `step` name.

    *Implementation hint:* Implement it to build a name for the given step using
    the class's custom formatting attributes. Since it is mainly meant for
    building names to save checkpoints, it can raise error if this NameFormat is
    just meant for finding already existing step paths.

    Args:
      step: Step number.
    """
    ...

  @abc.abstractmethod
  def find_all(self, base_path: epath.PathLike) -> Iterator[MetadataT]:
    """Returns metadata of all steps.

    NOTE: Ignores uncommitted checkpoints.

    *Implementation hint:* Implement it to find all step folders under
    `base_path` performing IO operations if needed. Use
    `build_step_metadatas(...)` helper function to build all the `MetadataT`
    using the found step paths.

    Args:
      base_path: *root* Path under which Step folders are placed.
    """
    ...

  @abc.abstractmethod
  def find_step(self, base_path: epath.PathLike, step: int) -> MetadataT:
    """Returns the metadata for `step` or raises ValueError.

    NOTE: Ignores uncommitted checkpoints.

    *Implementation hint:* Implement it to find the step folder under
    `base_path` performing IO operations if needed.

    Args:
      base_path: *root* Path under which Step folders are placed.
      step: Step number.
    """
    ...


def build_step_path(
    base_path: epath.PathLike, name_format: NameFormat[Metadata], step: int
) -> epath.Path:
  """Returns `step` path under `base_path` for step `name_format`."""
  return epath.Path(base_path) / name_format.build_name(step)


def build_step_metadatas(
    step_paths: Iterator[epath.Path],
    build_metadata: Callable[[epath.Path], Optional[MetadataT]],
) -> Iterator[MetadataT]:
  """Yields filtered metadata mapped with `step_paths`.

  Args:
    step_paths: Iterator of step paths.
    build_metadata: Callable to match and build step metadata from `step_paths`
      elements. If a `step_paths` element doesn't match then it returns None.

  Yields:
    Step metadata.
  """
  with concurrent.futures.ThreadPoolExecutor() as executor:
    metadata_futures = [
        executor.submit(build_metadata, step_path) for step_path in step_paths
    ]
    for future in concurrent.futures.as_completed(metadata_futures):
      metadata = future.result()
      if metadata is not None:
        yield metadata


def step_prefix_with_underscore(step_prefix: Optional[str]) -> str:
  """Returns `step_prefix` appended with `underscore` or <empty> if None."""
  return '' if step_prefix is None else f'{step_prefix}_'


def latest_step_metadata(
    root_path: epath.PathLike, name_format: NameFormat[MetadataT]
) -> Optional[MetadataT]:
  """Returns step.MetadataT of the latest step in `root_path`."""
  return max(
      sorted(
          name_format.find_all(root_path),
          key=lambda metadata: metadata.path.name,
          reverse=True,
      ),
      default=None,
      key=lambda metadata: metadata.step,
  )


def step_metadata_of_checkpoint_path(
    checkpoint_path: epath.PathLike, name_format: NameFormat[MetadataT]
) -> MetadataT:
  """Returns step.MetadataT of given `checkpoint_path`."""
  checkpoint_path = epath.Path(checkpoint_path)
  all_step_metadata = list(name_format.find_all(checkpoint_path.parent))
  for step_metadata in all_step_metadata:
    if step_metadata.path.name == checkpoint_path.name:
      return step_metadata
  raise ValueError(
      'Failed to resolve step metadata of checkpoint path with'
      f' NameFormat={name_format}, checkpoint path={checkpoint_path}, path'
      f' name({checkpoint_path.name}) did not match with available step names:'
      f' {[step_metadata.path.name for step_metadata in all_step_metadata]}.'
      ' Please check if the given path is really a checkpoint path.'
  )


# TODO(b/337858698): Works with CompositeNameFormat.write_name_format only. Also
# support read_name_formats.
def find_step_path(
    base_path: epath.PathLike,
    name_format: NameFormat[Metadata],
    *,
    step: int,
    include_uncommitted: bool = False,
) -> epath.Path:
  """Returns `step` path under `base_path` for step `name_format`.

  NOTE: Experimental function, subject to change.

  Args:
    base_path: directory path containing step subdirs.
    name_format: NameFormat of the target `step`.
    step: target step number.
    include_uncommitted: if True then uncommitted steps are considered in search
      too, otherwise only committed steps are looked up.

  Raises:
    ValueError if the target step path does not exist.
  """
  base_path = epath.Path(base_path)
  if not include_uncommitted:
    return name_format.find_step(base_path, step).path

  # First try finding uncommitted step.
  if is_gcs_path(base_path):
    uncommitted_step_path = build_step_path(base_path, name_format, step)
  else:
    step_name = name_format.build_name(step)
    uncommitted_step_path = None
    for uncommitted_name in tmp_checkpoints(base_path):
      if uncommitted_name.startswith(f'{step_name}{TMP_DIR_SUFFIX}'):
        uncommitted_step_path = base_path / uncommitted_name
        break
  if uncommitted_step_path and uncommitted_step_path.exists():
    return uncommitted_step_path
  # Uncommitted step not found, return committed one or raise error.
  return name_format.find_step(base_path, step).path


@dataclasses.dataclass(frozen=True)
class _StandardNameFormat(NameFormat[Metadata]):
  """NameFormat for 'standard' steps for common Orbax use cases.

  NOTE: Ignores uncommitted checkpoints.

  Naming examples:
   * step_prefix=None    step_format_fixed_length=None  ->  23
   * step_prefix=None    step_format_fixed_length=4     ->  0023
   * step_prefix=step    step_format_fixed_length=None  ->  step_23
   * step_prefix=step    step_format_fixed_length=4     ->  step_0023

  Attributes:
    step_prefix: Optional fixed string prefixed to step. Note an *underscore* is
      appended before applying it.
    step_format_fixed_length: Optional length of the zero padded step. e.g. 6
      for 000123.
  """

  step_prefix: Optional[str] = None
  step_format_fixed_length: Optional[int] = None

  def build_name(self, step: int) -> str:
    """Returns `(prefix_)?(zero padding)?step` name."""
    if self.step_format_fixed_length is not None:
      step_str = f'{step:0{self.step_format_fixed_length}d}'
    else:
      step_str = f'{step}'

    # [prefix]step
    return f'{step_prefix_with_underscore(self.step_prefix)}{step_str}'

  def _build_metadata(
      self, step_path: epath.Path, step: Optional[int] = None
  ) -> Optional[Metadata]:
    """Returns metadata for given `step_path` if it is valid or None."""
    if not step_path.is_dir():
      return None

    if not is_checkpoint_finalized(step_path):
      return None

    if step is not None:
      # step already known, just check exists.
      if step_path.exists():
        return Metadata(step=step, path=step_path)

    # Regex: [prefix]*(step)
    if self.step_format_fixed_length and self.step_format_fixed_length > 0:
      zero_present = rf'0\d{{{self.step_format_fixed_length-1}}}'
      zero_not_present = rf'[1-9]\d{{{self.step_format_fixed_length-1}}}\d*'
      zero_padded_step_group = rf'({zero_present}|{zero_not_present})'
    else:
      zero_padded_step_group = r'(0|[1-9]\d*)'
    name_regex = f'^{step_prefix_with_underscore(self.step_prefix)}{zero_padded_step_group}$'

    match = re.search(name_regex, step_path.name)
    if match is None:
      return None
    (step_,) = match.groups()
    step_ = int(step_)

    return Metadata(step=step_, path=step_path)

  def find_all(self, base_path: epath.PathLike) -> Iterator[Metadata]:
    """Returns metadata of all steps matching with name_format attributes."""
    # <step_prefix>_?<0 padding>?*
    step_paths = epath.Path(base_path).glob(
        f'{step_prefix_with_underscore(self.step_prefix)}*'
    )
    return build_step_metadatas(step_paths, self._build_metadata)

  def find_step(self, base_path: epath.PathLike, step: int) -> Metadata:
    """Returns the metadata for `step` or raises ValueError."""
    step_path = build_step_path(base_path, self, step)
    metadata = self._build_metadata(step_path, step=step)
    if metadata is not None:
      return metadata

    # Raise detailed error message.
    raise ValueError(
        f'No step path found with name={self.build_name(step)},'
        f' NameFormat={self} for step={step} under {base_path}.'
    )


def standard_name_format(
    step_prefix: Optional[str] = None,
    step_format_fixed_length: Optional[int] = None,
) -> NameFormat[Metadata]:
  """Returns NameFormat for 'standard' steps for common Orbax use cases.

  NOTE: Ignores uncommitted checkpoints.

  Naming examples:
   * step_prefix=None    step_format_fixed_length=None  ->  23
   * step_prefix=None    step_format_fixed_length=4     ->  0023
   * step_prefix=step    step_format_fixed_length=None  ->  step_23
   * step_prefix=step    step_format_fixed_length=4     ->  step_0023

  Args:
    step_prefix: Optional fixed string prefixed to step. Note an *underscore* is
      appended before applying it.
    step_format_fixed_length: Optional length of the zero padded step. e.g. 6
      for 000123.
  """
  return _StandardNameFormat(
      step_prefix=step_prefix,
      step_format_fixed_length=step_format_fixed_length,
  )


@dataclasses.dataclass(frozen=True)
class _CompositeNameFormat(NameFormat[Metadata]):
  """Supports reading multiple step namings, but just one format to write.

  Attributes:
    write_name_format: NameFormat used to build step names meant for writing
      checkpoints. Must be present in `read_name_formats` at a preferred
      priority position.
    read_name_formats: Sequence (ordered) of NameFormats used to find steps for
      reading checkpoints. It acts like an *or*, where the first one to match is
      returned.
  """

  write_name_format: NameFormat[Metadata]
  read_name_formats: Sequence[NameFormat[Metadata]]

  def __post_init__(self):
    if self.write_name_format not in self.read_name_formats:
      raise ValueError(
          f'write_name_format: {self.write_name_format} must be present in'
          f' read_name_formats: {self.read_name_formats}.'
      )

  def build_name(self, step: int) -> str:
    """Returns `step` name using `write_name_format`."""
    return self.write_name_format.build_name(step)

  def find_all(self, base_path: epath.PathLike) -> Iterator[Metadata]:
    """Returns metadata of all steps."""
    found_paths = set()
    for read_name_format in self.read_name_formats:
      for step_metadata in read_name_format.find_all(base_path):
        if step_metadata.path not in found_paths:
          found_paths.add(step_metadata.path)
          yield step_metadata

  def find_step(self, base_path: epath.PathLike, step: int) -> Metadata:
    """Returns the metadata for `step` or raises ValueError."""
    errors = []  # Used to raise the final collated error if needed.
    for read_name_format in self.read_name_formats:
      try:
        return read_name_format.find_step(base_path, step)
      except Exception as e:  # pylint: disable=broad-exception-caught
        logging.info(
            'Failed to find step=%s with NameFormat=%s under %s. Error: %s',
            step,
            read_name_format,
            base_path,
            e,
        )
        errors.append(e)

    # Raise the concatenated errors.
    messages = [f'{e}' for e in errors]
    raise ValueError('\n'.join(messages))


def composite_name_format(
    write_name_format: NameFormat[Metadata],
    read_name_formats: Sequence[NameFormat[Metadata]],
) -> NameFormat[Metadata]:
  """Returns *composite* NameFormat supporting multiple read/single write formats.

  Args:
    write_name_format: NameFormat used to build step names meant for writing
      checkpoints. Must be present in `read_name_formats` at a preferred
      priority position.
    read_name_formats: Sequence (ordered) of NameFormats used to find steps for
      reading checkpoints. Please note that to resolve conflicts (and avoid
      raising errors) in case of multiple NameFormats matching a given step, the
      sequence should be provided in highest to lowest priority order:
      NameFormat appearing earlier in the sequence is preferred.
  """
  return _CompositeNameFormat(write_name_format, read_name_formats)


# TODO(b/337137764) Can't move it to path/utils due to cyclic dependency.
# Explore other options.
def is_gcs_path(path: epath.Path) -> bool:
  return path.as_posix().startswith(_GCS_PATH_PREFIX)


# TODO(b/337137764) Can't move it to path/utils due to cyclic dependency.
# Explore other options.
def get_save_directory(
    step: int,
    directory: epath.PathLike,
    name: Optional[str] = None,
    step_prefix: Optional[str] = None,
    override_directory: Optional[epath.PathLike] = None,
    step_format_fixed_length: Optional[int] = None,
    step_name_format: Optional[NameFormat[Metadata]] = None,
) -> epath.Path:
  """Returns the standardized path to a save directory for a single item.

  Args:
    step: Step number.
    directory: Top level checkpoint directory.
    name: Item name ('params', 'state', 'dataset', etc.).
    step_prefix: Prefix applied to `step` (e.g. 'checkpoint').
    override_directory: If provided, step, directory, and step_prefix are
      ignored.
    step_format_fixed_length: Uses a fixed number of digits with leading zeros
      to represent the step number. If None, there are no leading zeros.
    step_name_format: NameFormat used to define step name for step and under
      given root directory. If provided, `step_prefix` and
      `step_format_fixed_length` are ignored.

  Returns:
    A directory.
  """
  if directory is None:
    raise ValueError('Directory cannot be None.')
  directory = epath.Path(directory)
  if override_directory is not None:
    result = epath.Path(override_directory)
  else:
    step_name_format = step_name_format or standard_name_format(
        step_prefix=step_prefix,
        step_format_fixed_length=step_format_fixed_length,
    )
    result = build_step_path(directory, step_name_format, step)
  if name is not None:
    result /= name
  return result


def is_tmp_checkpoint(path: epath.PathLike) -> bool:
  """Determines whether a directory is a tmp checkpoint path."""
  path = epath.Path(path)
  if not path.exists():
    raise ValueError(f'Path {path} does not exist.')
  if not path.is_dir():
    return False
  if is_gcs_path(path) and not (path / _COMMIT_SUCCESS_FILE).exists():
    return True
  if TMP_DIR_SUFFIX in path.name:
    return True
  return False


def is_checkpoint_finalized(path: epath.PathLike) -> bool:
  """Determines if the given path represents a finalized checkpoint.

  Path takes the form::

    path/to/my/dir/<name>.orbax-checkpoint-tmp-<timestamp>/  # not finalized
    path/to/my/dir/<name>/  # finalized

  Alternatively::

    gs://path/to/my/dir/<name>/  # finalized
      commit_success.txt
      ...
    gs://<path/to/my/dir/<name>/  # not finalized
      ...

  Args:
    path: Directory.

  Returns:
    True if the checkpoint is finalized.

  Raises:
    ValueError if the provided path is not a directory. Valid checkpoint paths
    must be a directory.
  """
  path = epath.Path(path)
  if not path.exists():
    raise ValueError(f'Path {path} does not exist.')
  if not path.is_dir():
    raise ValueError(f'Path {path} is not a directory. Not a valid checkpoint')
  if is_gcs_path(path) and not (path / _COMMIT_SUCCESS_FILE).exists():
    logging.warning(
        'This GCS path %s does not contain the %s file used to indicate a'
        ' successfully written GCS checkpoint. If the checkpoint was'
        ' originally saved with GCS, the checkpoint was not successfully'
        ' written. If the checkpoint was saved differently and copied, you'
        ' need to add %s to the checkpoint directory.',
        path,
        _COMMIT_SUCCESS_FILE,
        _COMMIT_SUCCESS_FILE,
    )

    return False
  if TMP_DIR_SUFFIX in path.name:
    return False
  return True


def tmp_checkpoints(checkpoint_dir: epath.PathLike) -> List[str]:
  """Returns a list of tmp checkpoint dir names in `checkpoint_dir`."""
  checkpoint_dir = epath.Path(checkpoint_dir)
  return [s.name for s in checkpoint_dir.iterdir() if is_tmp_checkpoint(s)]


def cleanup_tmp_directories(
    directory: epath.PathLike,
    primary_host: Optional[int] = 0,
    active_processes: Optional[Set[int]] = None,
    barrier_sync_key_prefix: Optional[str] = None,
):
  """Cleanup steps in `directory` with tmp files, as these are not finalized."""
  directory = epath.Path(directory)
  if multihost.is_primary_host(primary_host):
    logging.info('Cleaning up existing temporary directories at %s.', directory)
    tmp_files = tmp_checkpoints(directory)
    for tmp_file in tmp_files:
      (directory / tmp_file).rmtree()
  multihost.sync_global_processes(
      multihost.unique_barrier_key(
          'cleanup_tmp_dirs',
          prefix=barrier_sync_key_prefix,
          suffix=directory.name,
      ),
      timeout=multihost.DIRECTORY_DELETION_TIMEOUT,
      processes=active_processes,
  )


# TODO(b/337137764) Close to checkpoint finalization, but explore other options.
def record_saved_duration(checkpoint_start_time: float):
  """Record program duration that is accounted for by this checkpoint.

  For the very first checkpoint, this is the interval between program init and
  current checkpoint start time.

  Note that we use the checkpoint start time instead of end time. The saved
  duration should not include parallel training duration while the async
  checkpoint is being written in the background.

  Args:
    checkpoint_start_time: Start time of current checkpoint.
  """
  global _LAST_CHECKPOINT_WRITE_TIME
  # Note: for the very first checkpoint, this is the interval between program
  # init and the current checkpoint start time.
  duration_since_last_checkpoint = (
      checkpoint_start_time - _LAST_CHECKPOINT_WRITE_TIME
  )
  # TODO(hanyangtay): Remove version guard.
  if jax.version.__version_info__ > (0, 3, 25):
    jax.monitoring.record_event_duration_secs(
        '/jax/checkpoint/write/duration_since_last_checkpoint_secs',
        duration_since_last_checkpoint,
    )
  _LAST_CHECKPOINT_WRITE_TIME = checkpoint_start_time


def _is_step_checkpoint(path: epath.Path) -> bool:
  """Determines if the path resembles an Orbax step directory.

  Note that this is not foolproof, and users should not add extra files to the
  checkpoint directory beyond what is done by CheckpointManager.

  Args:
    path: path to check.

  Returns:
    bool indicating whether the path resembles an Orbax step directory.
  """
  name = path.name
  # Path must be a directory and either a digit, or end in '_' + digit.
  return path.is_dir() and (name.isdigit() or name.split('_')[-1].isdigit())


def step_from_checkpoint_name(name: str) -> int:
  """Returns the step from a checkpoint name. Also works for tmp checkpoints."""
  if name.isdigit():
    return int(name)
  elif name.split('_')[-1].isdigit():
    split = name.split('_')
    if len(split) == 2 and split[0]:
      return int(split[-1])
  elif tmp_match := re.match(TMP_DIR_STEP_PATTERN, name):
    return int(tmp_match.group(1))
  raise ValueError(f'Unrecognized name format: {name}.')


def checkpoint_steps_paths(
    checkpoint_dir: epath.PathLike,
) -> List[epath.Path]:
  """Returns a list of finalized checkpoint paths in the directory."""
  checkpoint_dir = epath.Path(checkpoint_dir)
  if not checkpoint_dir.exists():
    raise ValueError(f'Path {checkpoint_dir} does not exist.')

  def check_step_dir(step_dir: epath.Path) -> bool:
    # This block allows catching errors in which the checkpoint was deleted
    # between checking _is_step_checkpoint and is_checkpoint_finalized.
    try:
      result = _is_step_checkpoint(step_dir) and is_checkpoint_finalized(
          step_dir
      )
    except ValueError:
      return False

    return result

  with concurrent.futures.ThreadPoolExecutor() as executor:
    futures = {
        step_dir: executor.submit(check_step_dir, step_dir)
        for step_dir in checkpoint_dir.iterdir()
    }
    return [step_dir for step_dir, future in futures.items() if future.result()]


def checkpoint_steps(
    checkpoint_dir: epath.PathLike, single_host_load_and_broadcast: bool = False
) -> List[int]:
  """Returns a list of finalized checkpoint steps in the directory."""

  def _checkpoint_steps(path: epath.Path) -> List[int]:
    return [
        step_from_checkpoint_name(s.name) for s in checkpoint_steps_paths(path)
    ]

  checkpoint_dir = epath.Path(checkpoint_dir)
  if single_host_load_and_broadcast:
    max_steps = len(list(checkpoint_dir.iterdir()))
    # Read the step list only from host 0, and then broadcast the list.
    # This minimizes queries on non-leader processes.
    padded_step_list = np.array([-1] * max_steps)
    if multihost.process_index() == 0:
      steps = np.array(_checkpoint_steps(checkpoint_dir))
      assert len(steps) <= max_steps
      padded_step_list[0 : len(steps)] = steps
    padded_step_list = multihost.broadcast_one_to_all(padded_step_list)
    return [step for step in padded_step_list if step >= 0]
  return _checkpoint_steps(checkpoint_dir)


def any_checkpoint_step(checkpoint_dir: epath.PathLike) -> Optional[int]:
  """Returns any finalized checkpoint step in the directory or None.

  This avoids iterating over the entire directory.

  Args:
    checkpoint_dir: Checkpoint directory.

  Returns:
    Any finalized checkpoint step in the directory or None.
  """
  checkpoint_dir = epath.Path(checkpoint_dir)
  for s in checkpoint_dir.iterdir():
    if _is_step_checkpoint(s) and is_checkpoint_finalized(s):
      return step_from_checkpoint_name(s.name)
  return None



# Copyright 2024 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test for utils module."""

from typing import Any, Mapping, Sequence

from absl.testing import absltest
from absl.testing import parameterized
from etils import epath
import flax
import jax
import numpy as np
import optax
from orbax.checkpoint import test_utils
from orbax.checkpoint._src.testing import test_tree_utils
from orbax.checkpoint._src.tree import utils as tree_utils


# TODO: b/365169723 - Add tests: PyTreeMetadataOptions.support_rich_types=True.
class SerializeTreeTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self.directory = epath.Path(
        self.create_tempdir(name='serialize_tree_test').full_path
    )

  def test_serialize(self):
    tree = {'a': 1, 'b': {'c': {'d': 2}}, 'e': [1, {'x': 5, 'y': 7}, [9, 10]]}
    serialized = tree_utils.serialize_tree(tree, keep_empty_nodes=True)
    self.assertDictEqual(tree, serialized)
    deserialized = tree_utils.deserialize_tree(
        serialized, target=tree, keep_empty_nodes=True
    )
    test_utils.assert_tree_equal(self, tree, deserialized)

  def test_serialize_empty(self):
    serialized = tree_utils.serialize_tree({}, keep_empty_nodes=True)
    test_utils.assert_tree_equal(self, {}, serialized)
    deserialized = tree_utils.deserialize_tree(
        serialized, target={}, keep_empty_nodes=True
    )
    test_utils.assert_tree_equal(self, {}, deserialized)

  def test_serialize_empty_no_keep_empty(self):
    with self.assertRaises(ValueError):
      tree_utils.serialize_tree({}, keep_empty_nodes=False)

  def test_serialize_single_element(self):
    tree = {}
    serialized = tree_utils.serialize_tree(12345, keep_empty_nodes=True)
    test_utils.assert_tree_equal(self, 12345, serialized)
    deserialized = tree_utils.deserialize_tree(
        serialized, target=tree, keep_empty_nodes=True
    )
    test_utils.assert_tree_equal(self, 12345, deserialized)

  def test_serialize_list(self):
    tree = [1, {'a': 2}, [3, 4]]
    serialized = tree_utils.serialize_tree(tree, keep_empty_nodes=True)
    self.assertListEqual(tree, serialized)
    deserialized = tree_utils.deserialize_tree(
        serialized, target=tree, keep_empty_nodes=True
    )
    test_utils.assert_tree_equal(self, tree, deserialized)

  def test_serialize_filters_empty(self):
    tree = {'a': 1, 'b': None, 'c': {}, 'd': [], 'e': optax.EmptyState()}
    serialized = tree_utils.serialize_tree(tree, keep_empty_nodes=False)
    self.assertDictEqual({'a': 1}, serialized)
    deserialized = tree_utils.deserialize_tree(
        serialized, target=tree, keep_empty_nodes=False
    )
    test_utils.assert_tree_equal(self, tree, deserialized)

  def test_serialize_class(self):
    @flax.struct.dataclass
    class Foo(flax.struct.PyTreeNode):
      a: int
      b: Mapping[str, str]
      c: Sequence[optax.EmptyState]
      d: Sequence[Mapping[str, str]]

    foo = Foo(
        1,
        {'a': 'b', 'c': 'd'},
        [optax.EmptyState(), optax.EmptyState()],
        [{}, {'x': 'y'}, None],
    )
    serialized = tree_utils.serialize_tree(foo, keep_empty_nodes=True)
    expected = {
        'a': 1,
        'b': {'a': 'b', 'c': 'd'},
        'c': [optax.EmptyState(), optax.EmptyState()],
        'd': [{}, {'x': 'y'}, None],
    }
    self.assertDictEqual(expected, serialized)
    deserialized = tree_utils.deserialize_tree(
        serialized, target=foo, keep_empty_nodes=False
    )
    test_utils.assert_tree_equal(self, foo, deserialized)

  def test_serialize_nested_class(self):
    @flax.struct.dataclass
    class Foo(flax.struct.PyTreeNode):
      a: int

    nested = {
        'x': Foo(a=1),
        'y': {'z': Foo(a=2)},
    }
    serialized = tree_utils.serialize_tree(nested, keep_empty_nodes=True)
    expected = {
        'x': dict(a=1),
        'y': {'z': dict(a=2)},
    }
    self.assertDictEqual(expected, serialized)


class UtilsTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self.directory = epath.Path(
        self.create_tempdir(name='checkpointing_test').full_path
    )

  @parameterized.parameters(
      ({'a': 1, 'b': {'c': {}, 'd': 2}}, {('a',): 1, ('b', 'd'): 2}),
      ({'x': ['foo', 'bar']}, {('x', '0'): 'foo', ('x', '1'): 'bar'}),
  )
  def test_to_flat_dict(self, tree, expected):
    self.assertDictEqual(expected, tree_utils.to_flat_dict(tree))

  @parameterized.parameters(
      ({'a': 1, 'b': {'d': 2}}, {('a',): 1, ('b', 'd'): 2}),
      ({'x': ['foo', 'bar']}, {('x', '0'): 'foo', ('x', '1'): 'bar'}),
      ({'a': 1, 'b': 2}, {('b',): 2, ('a',): 1}),
  )
  def test_from_flat_dict(self, expected, flat_dict):
    empty = jax.tree.map(lambda _: 0, expected)
    self.assertDictEqual(
        expected, tree_utils.from_flat_dict(flat_dict, target=empty)
    )

  @parameterized.parameters(
      ({'a': 1, 'b': {'d': 2}}, {('a',): 1, ('b', 'd'): 2}),
      ({'a': 1, 'b': 2}, {('b',): 2, ('a',): 1}),
  )
  def test_from_flat_dict_without_target(self, expected, flat_dict):
    self.assertDictEqual(expected, tree_utils.from_flat_dict(flat_dict))

  @parameterized.parameters(
      ({'a': 1, 'b': {'d': 2}}, {'a': 1, 'b/d': 2}),
      ({'a': 1, 'b': 2}, {'b': 2, 'a': 1}),
      ({'a': {'b': {'c': 1}}}, {'a/b/c': 1}),
  )
  def test_from_flat_dict_with_sep(self, expected, flat_dict):
    self.assertDictEqual(
        expected, tree_utils.from_flat_dict(flat_dict, sep='/')
    )

  @parameterized.parameters(
      (1, True, False),
      (dict(), True, True),
      ({}, True, True),
      ({'a': {}}, False, False),
      ([], True, True),
      ([[]], False, False),
      ([tuple()], False, False),
      ([dict()], False, False),
      ([{}], False, False),
      ([1], False, False),
      (tuple(), True, True),
      ((tuple(),), False, False),
      (([],), False, False),
      (({},), False, False),
      ((dict(),), False, False),
      ((1,), False, False),
      (None, True, True),
      ((1, 2), False, False),
      (test_tree_utils.EmptyNamedTuple(), True, True),
      (test_tree_utils.MuNu(mu=None, nu=None), False, False),
      (test_tree_utils.MyEmptyClass(), True, False),
      (test_tree_utils.MyClass(), True, False),
      (test_tree_utils.MyClass(a=None, b=None), True, False),
      (test_tree_utils.MyClass(a=None, b=np.zeros(1)), True, False),
      (test_tree_utils.MyEmptyChex(), True, True),
      (test_tree_utils.MyChex(), False, False),
      (
          test_tree_utils.MyChex(my_jax_array=None, my_np_array=None),
          False,
          False,
      ),
      (
          test_tree_utils.MyChex(my_jax_array=None, my_np_array=np.zeros(1)),
          False,
          False,
      ),
  )
  def test_is_empty_or_leaf(
      self,
      value: Any,
      expected_is_empty_or_leaf: bool,
      expected_is_empty_node: bool,
  ):
    self.assertEqual(
        expected_is_empty_or_leaf, tree_utils.is_empty_or_leaf(value)
    )
    self.assertEqual(expected_is_empty_node, tree_utils.is_empty_node(value))


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from absl.testing import absltest
from absl.testing import parameterized
from orbax.checkpoint._src import composite as composite_lib


class CompositeTest(parameterized.TestCase):

  def test_composite_attributes(self):
    composite = composite_lib.Composite(a=1, b=2, d=4)
    self.assertEqual(1, composite.a)
    self.assertEqual(2, composite.b)
    self.assertEqual(4, composite.d)

  def test_composite_keys(self):
    composite = composite_lib.Composite(a=1, b=2, d=4)
    self.assertEqual({'a', 'b', 'd'}, composite.keys())

  def test_composite_values(self):
    composite = composite_lib.Composite(a=1, b=2, d=4)
    self.assertEqual({1, 2, 4}, set(composite.values()))

  def test_composite_items(self):
    composite = composite_lib.Composite(a=1, b=2, d=4)
    self.assertEqual({('a', 1), ('b', 2), ('d', 4)}, set(composite.items()))

  def test_composite_get(self):
    composite = composite_lib.Composite(a=1, b=2, d=4)
    self.assertEqual(1, composite['a'])
    self.assertEqual(2, composite['b'])
    self.assertEqual(4, composite['d'])

  def test_composite_len(self):
    composite = composite_lib.Composite(a=1, b=2, d=4)
    self.assertLen(composite, 3)

  def test_composite_get_invalid_key(self):
    composite = composite_lib.Composite(a=1, b=2, d=4)
    with self.assertRaises(KeyError):
      _ = composite['c']

  def test_composite_get_default(self):
    composite = composite_lib.Composite(a=1, b=2, d=4)
    self.assertIsNone(composite.get('c'))
    self.assertEqual(4, composite.get('c', 4))

  def test_invalid_key(self):
    with self.assertRaisesRegex(ValueError, 'cannot start with'):
      composite_lib.Composite(__invalid_name=2)

  def test_reserved_keys_are_unchanged(self):
    # To avoid breaking future users, make sure that Composite
    # only reserves the following attributes:
    reserved_keys = {'_is_protocol', '_abc_impl'}
    self.assertEqual(
        set([
            x for x in dir(composite_lib.Composite)
            if not x.startswith('__') and x not in reserved_keys
        ]),
        {'get', 'items', 'keys', 'values'},
    )

  def test_use_reserved_keys(self):
    composite = composite_lib.Composite(keys=3, values=4)

    self.assertNotEqual(3, composite.keys)
    self.assertNotEqual(4, composite.values)

    self.assertEqual(3, composite['keys'])
    self.assertEqual(4, composite['values'])

    self.assertEqual({'keys', 'values'}, composite.keys())
    self.assertEqual({3, 4}, set(composite.values()))

  def test_special_character_key(self):
    composite = composite_lib.Composite(**{
        '.invalid_attribute_but_valid_key': 15,
        'many special characters!': 16,
    })
    self.assertEqual(15, composite['.invalid_attribute_but_valid_key'])
    self.assertEqual(16, composite['many special characters!'])
    self.assertLen(composite, 2)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""To test Orbax in single-host setup."""

from absl.testing import absltest
from absl.testing import parameterized
from etils import epath
from orbax.checkpoint._src.path import deleter as deleter_lib
from orbax.checkpoint._src.path import step as step_lib


class CheckpointDeleterTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self.ckpt_dir = epath.Path(self.create_tempdir('ckpt').full_path)

  def _get_save_diretory(self, step: int, directory: epath.Path) -> epath.Path:
    return directory / str(step)

  @parameterized.product(
      threaded=(False, True),
      todelete_subdir=(None, 'some_delete_dir'),
  )
  def test_checkpoint_deleter_delete(self, threaded, todelete_subdir):
    """Test regular CheckpointDeleter."""
    deleter = deleter_lib.create_checkpoint_deleter(
        primary_host=None,
        directory=self.ckpt_dir,
        todelete_subdir=todelete_subdir,
        name_format=step_lib.standard_name_format(),
        enable_background_delete=threaded,
    )

    step = 1
    step_dir = self._get_save_diretory(step, self.ckpt_dir)
    step_dir.mkdir()
    self.assertTrue(step_dir.exists())
    deleter.delete(step)
    deleter.close()

    # assert the step_dir is deleted
    self.assertFalse(step_dir.exists())

    # In case of rename, check if the new folder exists
    if todelete_subdir is not None:
      self.assertTrue((self.ckpt_dir / todelete_subdir / str(step)).exists())

    deleter.close()


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from absl.testing import absltest
from absl.testing import parameterized
from etils import epath
import jax
import jax.numpy as jnp
import ml_dtypes
import numpy as np
from orbax.checkpoint import test_utils
from orbax.checkpoint._src.handlers import pytree_checkpoint_handler
from orbax.checkpoint._src.handlers import standard_checkpoint_handler_test_utils
from orbax.checkpoint._src.serialization import type_handlers
import tensorstore as ts


PyTreeCheckpointHandler = test_utils.PyTreeCheckpointHandler


class SingleHostTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self.ckpt_dir = epath.Path(self.create_tempdir('ckpt').full_path)

  def test_save_and_restore_a_single_device_sharded_jax_array(self):
    handler = PyTreeCheckpointHandler()
    key = jax.random.PRNGKey(0)
    x = jax.random.normal(key, (10,))
    assert isinstance(x.sharding, jax.sharding.SingleDeviceSharding)
    handler.save(
        self.ckpt_dir,
        args=pytree_checkpoint_handler.PyTreeSaveArgs({'array_x': x}),
    )

    restored_tree = handler.restore(self.ckpt_dir)
    np.testing.assert_array_equal(x, restored_tree['array_x'])

    self.assertIsInstance(restored_tree['array_x'], jax.Array)
    self.assertEqual(x.sharding, restored_tree['array_x'].sharding)

  @parameterized.parameters([False, True])
  def test_save_and_restore_jax_array(self, use_zarr3):
    handler = PyTreeCheckpointHandler(use_zarr3=use_zarr3)
    key = jax.random.PRNGKey(0)
    x = jax.random.normal(key, (10,))
    handler.save(self.ckpt_dir, {'x': x})
    restored_tree = handler.restore(self.ckpt_dir)

    np.testing.assert_array_equal(x, restored_tree['x'])
    assert isinstance(restored_tree['x'], jax.Array)

  def test_save_and_restore_zarrv3_jax_array_default_chunk_size(self):
    handler = PyTreeCheckpointHandler(use_zarr3=True)
    key = jax.random.PRNGKey(0)
    array_shape = (10, 7)
    x = jax.random.normal(key, array_shape)
    pytree = {'x': x}
    handler.save(self.ckpt_dir, pytree)

    # validate the stored array is in the chunk_layout specified
    tsstore = ts.open({
        'driver': 'zarr3',
        'kvstore': {
            'driver': 'ocdbt',
            'base': f'file://{self.ckpt_dir}',
            'path': 'x',
        },
    }).result()

    np.testing.assert_array_equal(
        tsstore.chunk_layout.read_chunk.shape, array_shape
    )
    np.testing.assert_array_equal(
        tsstore.chunk_layout.write_chunk.shape, array_shape
    )

    # validate the restored_tree is identical as the written one
    restore_handler = PyTreeCheckpointHandler(use_zarr3=True)
    restored_tree = restore_handler.restore(self.ckpt_dir)
    np.testing.assert_array_equal(x, restored_tree['x'])
    self.assertIsInstance(restored_tree['x'], jax.Array)

  def test_save_and_restore_zarrv3_with_metadata(self):
    handler = PyTreeCheckpointHandler(use_zarr3=True)
    key = jax.random.PRNGKey(0)
    x = jax.random.normal(key, (10,))
    pytree = {'x': x}
    handler.save(self.ckpt_dir, pytree)

    # even use_zarr3 is set to False, checkpoint can still be restored
    restore_handler = PyTreeCheckpointHandler(use_zarr3=False)
    restored_tree = restore_handler.restore(self.ckpt_dir)

    # validate the restored_tree is identical as the written one
    np.testing.assert_array_equal(x, restored_tree['x'])
    self.assertIsInstance(restored_tree['x'], jax.Array)

  # TODO: b/354139177 - Add a test for invalid chunk_byte_size values.
  def test_chunk_byte_size(self):
    handler = PyTreeCheckpointHandler(use_zarr3=True)
    key = jax.random.PRNGKey(0)
    x = jax.random.normal(key, (10, 100, 200), dtype=jnp.dtype('float32'))
    pytree = {'x': x}

    save_args = jax.tree.map(
        lambda x: type_handlers.SaveArgs(
            chunk_byte_size=jnp.dtype('float32').itemsize * 5**3,
        ),
        pytree,
    )
    handler.save(self.ckpt_dir, pytree, save_args=save_args)

    # validate the stored array is in the chunk_layout specified
    tsstore = ts.open({
        'driver': 'zarr3',
        'kvstore': {
            'driver': 'ocdbt',
            'base': f'file://{self.ckpt_dir}',
            'path': 'x',
        },
    }).result()

    np.testing.assert_array_equal(
        tsstore.chunk_layout.read_chunk.shape, (5, 5, 5)
    )
    np.testing.assert_array_equal(
        tsstore.chunk_layout.write_chunk.shape, (5, 5, 5)
    )

    # validate the restored_tree is identical as the written one
    restore_handler = PyTreeCheckpointHandler(use_zarr3=True)
    restored_tree = restore_handler.restore(self.ckpt_dir)
    np.testing.assert_array_equal(x, restored_tree['x'])
    assert isinstance(restored_tree['x'], jax.Array)

  @parameterized.parameters({'x': [1, 2]}, {'x': 1})
  def test_save_singular_array_with_standard_checkpoint_handler(self, x):
    if isinstance(x, list):
      x = jnp.array(x)
    handler = standard_checkpoint_handler_test_utils.StandardCheckpointHandler()
    with self.assertRaisesRegex(
        ValueError, '.*Use ArrayCheckpointHandler / ArraySave.*'
    ):
      handler.save(
          self.ckpt_dir,
          args=standard_checkpoint_handler_test_utils.StandardSaveArgs(x),
      )

  @parameterized.product(
      dtype=[
          jnp.bfloat16,
          ml_dtypes.bfloat16,
          np.dtype('float32'),
          np.dtype('float64'),
          ml_dtypes.float8_e5m2fnuz,
          ml_dtypes.float8_e5m2,
          ml_dtypes.float8_e4m3fnuz,
          ml_dtypes.float8_e4m3fn,
          ml_dtypes.float8_e4m3b11fnuz,
          ml_dtypes.int4,
          # ml_dtypes.uint4, # TODO(b/295577703)
          jnp.uint8,
          jnp.uint16,
          jnp.uint32,
          jnp.uint64,
          jnp.int8,
          jnp.int16,
          jnp.int32,
          jnp.int64,
      ],
      use_ocdbt=[True, False],
      use_zarr3=[True, False],
  )
  def test_save_and_restore_different_dtypes(self, dtype, use_ocdbt, use_zarr3):
    handler = PyTreeCheckpointHandler(use_ocdbt=use_ocdbt, use_zarr3=use_zarr3)

    rand_arr = np.random.normal(scale=3, size=10)  # use int4 max
    x = jax.numpy.asarray(rand_arr, dtype=dtype)
    pytree = {'x': x}
    handler.save(self.ckpt_dir, pytree)

    restore_handler = PyTreeCheckpointHandler()
    restored_tree = restore_handler.restore(self.ckpt_dir)

    # validate the restored_tree is identical as the written one
    np.testing.assert_array_equal(x, restored_tree['x'])
    self.assertIsInstance(restored_tree['x'], jax.Array)

  def test_save_and_restore_with_replica_parallel(self):
    if len(jax.devices()) <= 1:
      self.skipTest('Test requires multiple devices.')
    mesh = jax.sharding.Mesh(jax.devices(), ('x',))
    replicated_spec = jax.sharding.PartitionSpec()
    sharding = jax.sharding.NamedSharding(mesh, replicated_spec)

    key = jax.random.PRNGKey(0)
    state = jax.random.normal(key, (1024, 1024))
    state = jax.device_put(state, sharding)

    pytree = {'state': state}
    array_handler = type_handlers.ArrayHandler(
        replica_id=0,
        use_replica_parallel=True,
    )
    handler = PyTreeCheckpointHandler(
        type_handler_registry=type_handlers.create_type_handler_registry(
            (jax.Array, array_handler),
        ),
    )
    handler.save(self.ckpt_dir, pytree)

    restored_tree = handler.restore(self.ckpt_dir)

    np.testing.assert_array_equal(state, restored_tree['state'])
    self.assertIsInstance(restored_tree['state'], jax.Array)


if __name__ == '__main__':
  absltest.main()
